from contextlib import nullcontext as does_not_raise
from contextlib import suppress
from itertools import product

import numpy as np
import pyrtools as pt
import pytest
import torch

import plenoptic as po
from conftest import DEVICE, IMG_DIR
from plenoptic.tools.data import to_numpy

ALL_SPYRS = [
    f"{h}-{o}-{c}-{d}-{tf}"
    for h, o, c, d, tf in product(
        ["auto", 1, 3, 4, 5],
        [1, 2, 3],
        [True, False],
        [True, False],
        [True, False],
    )
] + [
    # pyramid with order=0 can only be non-complex
    f"{h}-0-False-{d}-{tf}"
    for h, d, tf in product(
        ["auto", 1, 3, 4, 5],
        [True, False],
        [True, False],
    )
]


def check_pyr_coeffs(coeff_1, coeff_2, rtol=1e-3, atol=1e-3):
    """
    function that checks if two sets of pyramid coefficients are the same
    We set an absolute and relative tolerance and the following function checks if
    abs(coeff1-coeff2) <= atol + rtol*abs(coeff1)
    Inputs:
    coeff1: first dictionary of pyramid coefficients
    coeff2: second dictionary of pyramid coefficients
    Both coeffs must obviously have the same number of scales, orientations etc.
    """

    for k in coeff_1:
        if torch.is_tensor(coeff_1[k]):
            coeff_1_np = to_numpy(coeff_1[k].squeeze())
        else:
            coeff_1_np = coeff_1[k]
        if torch.is_tensor(coeff_2[k]):
            coeff_2_np = to_numpy(coeff_2[k].squeeze())
        else:
            coeff_2_np = coeff_2[k]

        np.testing.assert_allclose(coeff_1_np, coeff_2_np, rtol=rtol, atol=atol)


def check_band_energies(coeff_1, coeff_2, rtol=1e-4, atol=1e-4):
    """
    function that checks if the energy in each band of two pyramids are the same.
    We set an absolute and relative tolerance and the function checks for each band if
    abs(coeff_1-coeff_2) <= atol + rtol*abs(coeff_1)
    Args:
    coeff_1: first dictionary of torch tensors corresponding to each band
    coeff_2: second dictionary of torch tensors corresponding to each band
    """

    for i in range(len(coeff_1.items())):
        k1 = list(coeff_1.keys())[i]
        k2 = list(coeff_2.keys())[i]
        band_1 = to_numpy(coeff_1[k1])
        band_2 = to_numpy(coeff_2[k2])
        band_1 = band_1.squeeze()
        band_2 = band_2.squeeze()

        np.testing.assert_allclose(
            np.sum(np.abs(band_1) ** 2),
            np.sum(np.abs(band_2) ** 2),
            rtol=rtol,
            atol=atol,
        )


def check_parseval(im, coeff, rtol=1e-4, atol=0):
    """
    function that checks if the pyramid is parseval, i.e. energy of coeffs is
    the same as the energy in the original image.
    Args:
    input image: image stimulus as ``torch.Tensor``
    coeff: dictionary of torch tensors corresponding to each band
    """
    total_band_energy = 0
    im_energy = np.sum(to_numpy(im) ** 2)
    for k, v in coeff.items():
        band = to_numpy(coeff[k])
        band = band.squeeze()

        total_band_energy += np.sum(np.abs(band) ** 2)

    np.testing.assert_allclose(total_band_energy, im_energy, rtol=rtol, atol=atol)


class TestSteerablePyramid:
    @pytest.fixture(
        scope="class",
        params=[
            f"{im}-{shape}"
            for im in ["einstein", "curie"]
            for shape in [None, 224, "128_1", "128_2"]
        ],
    )
    def img(self, request):
        im, shape = request.param.split("-")
        img = po.load_images(IMG_DIR / "256" / f"{im}.pgm").to(DEVICE)
        if shape == "224":
            img = img[..., :224, :224]
        elif shape == "128_1":
            img = img[..., :128, :]
        elif shape == "128_2":
            img = img[..., :128]
        return img

    @pytest.fixture(
        scope="class",
        params=[f"{shape}" for shape in [None, 224, "128_1", "128_2"]],
    )
    def multichannel_img(self, request):
        shape = request.param
        # use fixture for img and use color_wheel instead.
        img = po.load_images(IMG_DIR / "mixed" / "flowers.jpg", as_gray=False).to(
            DEVICE
        )
        if shape == "224":
            img = img[..., :224, :224]
        elif shape == "128_1":
            img = img[..., :128, :]
        elif shape == "128_2":
            img = img[..., :128]
        return img

    # WARNING: because this fixture requires the img fixture above, it should
    # only be used in tests that also use the img fixture. That is, tests where
    # you want to test both the einstein and curie images, as well as the
    # different sizes. Otherwise, this will generate a bunch of tests that use
    # the spyr with those strange shapes
    @pytest.fixture(scope="class")
    def spyr(self, img, request):
        height, order, is_complex, downsample, tightframe = request.param.split("-")

        with suppress(ValueError):
            # then height = 'auto' and that's fine
            height = int(height)
        # need to use eval to get from 'False' (string) to False (bool);
        # bool('False') == True, annoyingly enough
        pyr = po.simul.SteerablePyramidFreq(
            img.shape[-2:],
            height,
            int(order),
            is_complex=eval(is_complex),
            downsample=eval(downsample),
            tight_frame=eval(tightframe),
        )
        pyr.to(DEVICE)
        return pyr

    @pytest.fixture(scope="class")
    def spyr_multi(self, multichannel_img, request):
        height, order, is_complex, downsample, tightframe = request.param.split("-")

        with suppress(ValueError):
            # then height = 'auto' and that's fine
            height = int(height)
        # need to use eval to get from 'False' (string) to False (bool);
        # bool('False') == True, annoyingly enough
        pyr = po.simul.SteerablePyramidFreq(
            multichannel_img.shape[-2:],
            height,
            int(order),
            is_complex=eval(is_complex),
            downsample=eval(downsample),
            tight_frame=eval(tightframe),
        )
        pyr.to(DEVICE)
        return pyr

    # can't use one of the spyr fixtures here because we need to instantiate separately
    # for each of these shapes
    @pytest.mark.parametrize("height", ["auto", 1, 3, 4, 5])
    @pytest.mark.parametrize("order", [0, 1, 2, 3])
    @pytest.mark.parametrize("is_complex", [True, False])
    @pytest.mark.parametrize(
        "im_shape",
        [None, (255, 255), (256, 128), (128, 256), (255, 256), (256, 255)],
    )
    @pytest.mark.filterwarnings(
        "ignore:Reconstruction will not be perfect with odd-sized images:UserWarning"
    )
    def test_pyramid(self, basic_stim, height, order, is_complex, im_shape):
        if im_shape is not None:
            basic_stim = basic_stim[..., : im_shape[0], : im_shape[1]]
        expectation = does_not_raise()
        if order == 0 and is_complex:
            expectation = pytest.raises(
                ValueError, match="Complex pyramid cannot have order=0"
            )
        elif im_shape is not None and any([im_shape[0] % 2, im_shape[1] % 2]):
            expectation = pytest.warns(
                Warning, match="Reconstruction will not be perfect"
            )
        with expectation:
            spc = po.simul.SteerablePyramidFreq(
                basic_stim.shape[-2:],
                height=height,
                order=order,
                is_complex=is_complex,
            ).to(DEVICE)
            spc(basic_stim)

    @pytest.mark.parametrize(
        "spyr",
        [
            f"{h}-{o}-{c}-{d}-True"
            for h, o, c, d in product(
                ["auto", 1, 2, 3], [1, 2, 3], [True, False], [True, False]
            )
        ]
        + [
            # pyramid with order=0 can only be non-complex
            f"{h}-0-False-{d}-True"
            for h, d in product(["auto", 1, 2, 3], [True, False])
        ],
        indirect=True,
    )
    def test_tight_frame(self, img, spyr):
        pyr_coeffs = spyr.forward(img)
        check_parseval(img, pyr_coeffs)

    @pytest.mark.parametrize(
        "spyr",
        [
            f"{h}-{o}-{c}-True-{t}"
            for h, o, c, t in product(
                [3, 4, 5], [1, 2, 3], [True, False], [True, False]
            )
        ]
        + [
            # pyramid with order=0 can only be non-complex
            f"{h}-0-False-True-{t}"
            for h, t in product([3, 4, 5], [True, False])
        ],
        indirect=True,
    )
    def test_not_downsample(self, img, spyr):
        pyr_coeffs = spyr.forward(img)
        # need to add 1 because our heights are 0-indexed (i.e., the lowest
        # height has k[0]==0)
        height = max([k[0] for k in pyr_coeffs if isinstance(k[0], int)]) + 1
        # couldn't come up with a way to get this with fixtures, so we
        # instantiate it each time.
        spyr_not_downsample = po.simul.SteerablePyramidFreq(
            img.shape[-2:],
            height,
            spyr.order,
            is_complex=spyr.is_complex,
            downsample=False,
            tight_frame=spyr.tight_frame,
        )
        spyr_not_downsample.to(DEVICE)
        pyr_coeffs_nd = spyr_not_downsample.forward(img)
        check_band_energies(pyr_coeffs, pyr_coeffs_nd)

    @pytest.mark.parametrize(
        "scales",
        [
            [0],
            [1],
            [0, 1, 2],
            [2],
            None,
            ["residual_highpass", "residual_lowpass"],
            ["residual_highpass", 0, 1, "residual_lowpass"],
        ],
    )
    @pytest.mark.parametrize(
        "spyr",
        [
            f"{h}-{o}-{c}-False-False"
            for h, o, c in product([3, 4, 5], [1, 2, 3], [True, False])
        ]
        + [
            # pyramid with order=0 can only be non-complex
            f"{h}-0-False-False-False"
            for h in [3, 4, 5]
        ],
        indirect=True,
    )
    def test_pyr_to_tensor(self, img, spyr, scales, rtol=1e-12, atol=1e-12):
        pyr_coeff_dict = spyr.forward(img, scales=scales)
        split_complex = [True, False] if spyr.is_complex else [False]

        for val in split_complex:
            pyr_tensor, pyr_info = spyr.convert_pyr_to_tensor(
                pyr_coeff_dict, split_complex=val
            )
            pyr_coeff_dict2 = spyr.convert_tensor_to_pyr(pyr_tensor, *pyr_info)
            check_pyr_coeffs(pyr_coeff_dict, pyr_coeff_dict2, rtol, atol)

    @pytest.mark.parametrize(
        "spyr",
        ALL_SPYRS,
        indirect=True,
    )
    def test_residuals_real_valued(self, img, spyr):
        pyr_coeff_dict = spyr.forward(img)
        if pyr_coeff_dict["residual_highpass"].dtype.is_complex:
            raise ValueError("Somehow residual_highpass is complex!")
        if pyr_coeff_dict["residual_lowpass"].dtype.is_complex:
            raise ValueError("Somehow residual_lowpass is complex!")
        if (not spyr.downsample) and spyr.is_complex:
            pyr_tensor, pyr_info = spyr.convert_pyr_to_tensor(
                pyr_coeff_dict,
                split_complex=False,
            )
            # residuals get cast to complex as part of the conversion to tensor (if the
            # pyramid is complex and we set split_complex=False), but still should have
            # no imaginary component
            if not torch.equal(
                pyr_tensor[:, 0].imag, torch.zeros_like(pyr_tensor[:, 0])
            ):
                raise ValueError(
                    "Somehow residual_highpass has non-zero imaginary values after"
                    " pyr_to_tensor!"
                )
            if not torch.equal(
                pyr_tensor[:, -1].imag, torch.zeros_like(pyr_tensor[:, -1])
            ):
                raise ValueError(
                    "Somehow residual_lowpass has non-zero imaginary values after "
                    "pyr_to_tensor!"
                )

    @pytest.mark.parametrize(
        "spyr",
        [
            f"{h}-{o}-{c}-True-False"
            for h, o, c in product([3, 4, 5], [1, 2, 3], [True, False])
        ]
        + [
            # pyramid with order=0 can only be non-complex
            f"{h}-0-False-True-False"
            for h in [3, 4, 5]
        ],
        indirect=True,
    )
    def test_torch_vs_numpy_pyr(self, img, spyr):
        torch_spc = spyr.forward(img)
        # need to add 1 because our heights are 0-indexed (i.e., the lowest
        # height has k[0]==0)
        height = max([k[0] for k in torch_spc if isinstance(k[0], int)]) + 1
        pyrtools_sp = pt.pyramids.SteerablePyramidFreq(
            to_numpy(img.squeeze()),
            height=height,
            order=spyr.order,
            is_complex=spyr.is_complex,
        )
        pyrtools_spc = pyrtools_sp.pyr_coeffs
        check_pyr_coeffs(pyrtools_spc, torch_spc)

    @pytest.mark.parametrize(
        "spyr",
        ALL_SPYRS,
        indirect=True,
    )
    def test_complete_recon(self, img, spyr):
        pyr_coeffs = spyr.forward(img)
        recon = to_numpy(spyr.recon_pyr(pyr_coeffs))
        np.testing.assert_allclose(recon, to_numpy(img), rtol=1e-4, atol=1e-4)

    @pytest.mark.parametrize(
        "spyr_multi",
        ALL_SPYRS,
        indirect=True,
    )
    def test_complete_recon_multi(self, multichannel_img, spyr_multi):
        pyr_coeffs = spyr_multi.forward(multichannel_img)
        recon = to_numpy(spyr_multi.recon_pyr(pyr_coeffs))
        np.testing.assert_allclose(
            recon, to_numpy(multichannel_img), rtol=1e-4, atol=1e-4
        )

    @pytest.mark.parametrize(
        "spyr",
        [
            f"{h}-{o}-{c}-{d}-{tf}"
            for h, o, c, d, tf in product(
                ["auto"], [3], [True, False], [True, False], [True, False]
            )
        ],
        indirect=True,
    )
    def test_partial_recon(self, img, spyr):
        pyr_coeffs = spyr.forward(img)
        # need to add 1 because our heights are 0-indexed (i.e., the lowest
        # height has k[0]==0)
        height = max([k[0] for k in pyr_coeffs if isinstance(k[0], int)]) + 1
        pt_spyr = pt.pyramids.SteerablePyramidFreq(
            to_numpy(img.squeeze()),
            height=height,
            order=spyr.order,
            is_complex=spyr.is_complex,
        )
        recon_levels = [[0], [1, 3], [1, 3, 4]]
        recon_bands = [[1], [1, 3]]
        for levels, bands in product(["all"] + recon_levels, ["all"] + recon_bands):
            po_recon = to_numpy(spyr.recon_pyr(pyr_coeffs, levels, bands).squeeze())
            pt_recon = pt_spyr.recon_pyr(levels, bands)
            np.testing.assert_allclose(po_recon, pt_recon, rtol=1e-4, atol=1e-4)

    @pytest.mark.parametrize(
        "spyr",
        [
            f"{h}-{o}-{c}-True-False"
            for h, o, c in product(["auto", 1, 3, 4], [1, 2, 3], [True, False])
        ]
        + [
            # pyramid with order=0 can only be non-complex
            f"{h}-0-False-True-False"
            for h in [3, 4, 5]
        ],
        indirect=True,
    )
    def test_recon_match_pyrtools(self, img, spyr, rtol=1e-6, atol=1e-6):
        # this should fail if and only if test_complete_recon does, but
        # may as well include it just in case
        pyr_coeffs = spyr.forward(img)
        # need to add 1 because our heights are 0-indexed (i.e., the lowest
        # height has k[0]==0)
        height = max([k[0] for k in pyr_coeffs if isinstance(k[0], int)]) + 1
        pt_pyr = pt.pyramids.SteerablePyramidFreq(
            to_numpy(img.squeeze()),
            height=height,
            order=spyr.order,
            is_complex=spyr.is_complex,
        )
        po_recon = po.to_numpy(spyr.recon_pyr(pyr_coeffs).squeeze())
        pt_recon = pt_pyr.recon_pyr()
        np.testing.assert_allclose(po_recon, pt_recon, rtol=rtol, atol=atol)

    @pytest.mark.parametrize(
        "scales",
        [
            [0],
            [4],
            [0, 1, 2],
            [0, 3, 4],
            ["residual_highpass", "residual_lowpass"],
            ["residual_highpass", 0, 1, "residual_lowpass"],
        ],
    )
    @pytest.mark.parametrize(
        "spyr",
        [f"auto-3-{c}-{d}-False" for c, d in product([True, False], [True, False])],
        indirect=True,
    )
    def test_scales_arg(self, img, spyr, scales):
        pyr_coeffs = spyr.forward(img)
        reduced_pyr_coeffs = spyr.forward(img, scales)
        for k, v in reduced_pyr_coeffs.items():
            if (v != pyr_coeffs[k]).any():
                raise Exception(
                    "Reduced pyr_coeffs should be same as original, but at"
                    f" least key {k} is not"
                )

        # recon_pyr should always fail
        with pytest.raises(Exception):
            spyr.recon_pyr()
        with pytest.raises(Exception):
            spyr.recon_pyr(scales)

    @pytest.mark.parametrize("height", range(-1, 8))
    def test_height_values(self, img, height):
        if height < 0:
            expectation = pytest.raises(
                ValueError, match="Height must be a non-negative int"
            )
        elif height > np.log2(min(img.shape[-2:])) - 2:
            expectation = pytest.raises(
                ValueError, match="Cannot build pyramid higher than"
            )
        else:
            expectation = does_not_raise()
        with expectation:
            pyr = po.simul.SteerablePyramidFreq(img.shape[-2:], height=height).to(
                DEVICE
            )
            pyr(img)

    @pytest.mark.parametrize("order", range(-1, 17))
    def test_order_values(self, img, order):
        if order in [-1, 16]:
            expectation = pytest.raises(
                ValueError, match="order must be an integer in the range"
            )
        else:
            expectation = does_not_raise()
        with expectation:
            pyr = po.simul.SteerablePyramidFreq(img.shape[-2:], order=order).to(DEVICE)
            pyr(img)

    @pytest.mark.parametrize("order", range(0, 16))
    def test_buffers(self, order):
        pyr = po.simul.SteerablePyramidFreq((256, 256), order=order)
        buffers = [k for k, _ in pyr.named_buffers()]
        names = ["lo0mask", "hi0mask"]
        for s in range(pyr.num_scales):
            names.extend(
                [
                    f"_himasks_scale_{s}",
                    f"_lomasks_scale_{s}",
                    f"_anglemasks_scale_{s}",
                    f"_anglemasks_recon_scale_{s}",
                ]
            )
        assert len(buffers) == len(names), (
            "pyramid doesn't have the right number of buffers!"
        )
        assert set(buffers) == set(names), "pyramid doesn't have the right buffers!"

    @pytest.mark.filterwarnings(
        "ignore:Reconstruction will not be perfect with odd-sized images:UserWarning"
    )
    def test_img_shape_error(self, curie_img):
        pyr = po.simul.SteerablePyramidFreq((255, 255)).to(DEVICE)
        with pytest.raises(
            ValueError, match="Input tensor height/width.*does not match"
        ):
            pyr(curie_img)

    @pytest.mark.parametrize("shape_type", ["tuple", "list", "torch.Size", "floats"])
    def test_img_shape_cast(self, curie_img, shape_type):
        # image shape will most likely be passed as tuple, list, or torch.Size objects,
        # all of those should work. also, if they're floats that can be parsed as ints.
        if shape_type == "tuple":
            shape = (256, 256)
        elif shape_type == "list":
            shape = [256, 256]
        elif shape_type == "torch.Size":
            shape = curie_img.shape[-2:]
        elif shape_type == "floats":
            shape = (256.0, 256.0)
        pyr = po.simul.SteerablePyramidFreq(shape).to(DEVICE)
        pyr(curie_img)

    @pytest.mark.parametrize(
        "error_type", ["length", "length_tensor", "floats", "other_dtype"]
    )
    def test_img_shape_error_init(self, curie_img, error_type):
        if error_type == "length":
            shape = (1, 1, 256, 256)
            expect_str = "image_shape must be a tuple of length 2"
        elif error_type == "length_tensor":
            shape = curie_img.shape
            expect_str = "image_shape must be a tuple of length 2"
        elif error_type == "floats":
            shape = (256.5, 256.6)
            expect_str = "image_shape must be castable to ints"
        elif error_type == "other_dtype":
            shape = ("hi", "there")
            expect_str = "image_shape must be castable to ints"
        with pytest.raises(ValueError, match=expect_str):
            po.simul.SteerablePyramidFreq(shape)

    @pytest.mark.parametrize(
        "spyr",
        [
            f"{h}-{o}-{c}-{d}-{tf}"
            for h, o, c, d, tf in product(
                ["auto", 1, 2, 3],
                [1, 2, 3],
                [True, False],
                [True, False],
                [True, False],
            )
        ]
        + [
            # pyramid with order=0 can only be non-complex
            f"{h}-0-False-{d}-{tf}"
            for h, d, tf in product(["auto", 1, 2, 3], [True, False], [True, False])
        ],
        indirect=True,
    )
    def test_steer_coeffs(self, img, spyr):
        pyr_coeffs = spyr(img)
        if spyr.is_complex:
            expectation = pytest.raises(
                AssertionError, match="steering only implemented"
            )
        else:
            expectation = does_not_raise()
        with expectation:
            spyr.steer_coeffs(pyr_coeffs, torch.linspace(0, np.pi, 32))



"""
Model metrics.

Simple functions to convert models, which can return a tensor of arbitrary shape, to
metrics, which must return a tensor.
"""

import torch


def model_metric(
    x: torch.Tensor, y: torch.Tensor, model: torch.nn.Module
) -> torch.Tensor:
    r"""
    Calculate distance between x and y in model space root mean squared error.

    For two images, :math:`x` and :math:`y`, and model :math:`M`.

    .. math::

        metric = \sqrt{\frac{1}{n}\sum_i (M(x)_i - M(y)_i)^2+\epsilon}

    where :math:`M(x)` and :math:`M(y)` are the model representations of ``x`` and
    ``y``, with :math:`n` elements, and :math:`\epsilon=1e-10` is to stabilize the
    gradient around zero.

    Parameters
    ----------
    x, y
        Images to pass to ``model``.
    model
        Torch model with defined forward operation.

    Returns
    -------
    model_error
        Root mean-squared error between the model representation of ``x`` and ``y``.

    Examples
    --------
    >>> import plenoptic as po
    >>> einstein_img = po.data.einstein()
    >>> curie_img = po.data.curie()
    >>> model = po.simul.Gaussian(30)
    >>> model_metric = po.metric.model_metric(einstein_img, curie_img, model)
    >>> model_metric
    tensor(0.3128, grad_fn=<SqrtBackward0>)
    >>> # calculate this model metric manually:
    >>> torch.mean((model(einstein_img) - model(curie_img)).pow(2)).sqrt()
    tensor(0.3128, grad_fn=<SqrtBackward0>)
    """
    repx = model(x)
    repy = model(y)

    # for optimization purpose (stabilizing the gradient around zero)
    epsilon = 1e-10

    dist = torch.sqrt(torch.mean((repx - repy) ** 2) + epsilon)

    return dist



#!/usr/bin/env python3

import pathlib
import re
import sys

failures = []

paths = []
for p in sys.argv[1:]:
    p = pathlib.Path(p)
    if p.is_dir():
        p = list(p.glob("**/*.py"))
    elif p.suffix == ".py":
        p = [p]
    else:
        p = []
    paths.extend(p)

for p in paths:
    with open(p) as f:
        contents = f.read()
    fail = re.findall(r"\.\.[a-zA-Z]+::", contents)
    if len(fail):
        failures.append((p, fail))

if len(failures):
    print("Found the following malformed sphinx directives! Add a space after the dots")
    for p, fail in failures:
        fail = ", ".join(fail)
        print(f"{p}: {fail}")
    sys.exit(1)



"""
Metrics designed to model human perceptual distance.

Metrics that model human perceptual distance seek to answer the question "how different
do humans find these two images?".
"""

import warnings
from importlib import resources
from pathlib import Path
from typing import Literal

import numpy as np
import torch
import torch.nn.functional as F

from ..simulate.canonical_computations import LaplacianPyramid
from ..simulate.canonical_computations.filters import circular_gaussian2d
from ..tools.conv import same_padding

DIRNAME = resources.files("plenoptic.metric")


def _ssim_parts(
    img1: torch.Tensor,
    img2: torch.Tensor,
    pad: Literal[False, "constant", "reflect", "replicate", "circular"] = False,
    func_name: str = "SSIM",
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Calculate the various components used to compute SSIM.

    This should not be called by users directly, but is meant to assist for
    calculating SSIM and MS-SSIM.

    Parameters
    ----------
    img1
        The first image or batch of images, of shape (batch, channel, height, width).
    img2
        The second image or batch of images, of shape (batch, channel, height, width).
        The heights and widths of ``img1`` and ``img2`` must be the same. The numbers of
        batches and channels of ``img1`` and ``img2`` need to be broadcastable: either
        they are the same or one of them is 1. The output will be computed separately
        for each channel (so channels are treated in the same way as batches). Both
        images should have values between 0 and 1. Otherwise, the result may be
        inaccurate, and we will raise a warning (but will still compute it).
    pad
        If not ``False``, how to pad the image for the convolutions computing the
        local average of each image. See :func:`torch.nn.functional.pad` for how
        these work.
    func_name
        Name of the function that called this one, in order to raise more helpful error
        / warning messages.

    Returns
    -------
    map_ssim
        Map of SSIM values across the image.
    contrast_structure_map
        Map of contrast structure values.
    weight
        Weight used for stability of computation.

    Raises
    ------
    ValueError
        If either ``img1`` or ``img2`` is not 4d.
    ValueError
        If ``img1`` and ``img2`` have different height or width.
    ValueError
        If ``img1`` and ``img2`` have different batch or channel, unless one of them has
        a 1 there, so they can be broadcast.
    ValueError
        If ``img1`` and ``img2`` have different dtypes.

    Warns
    -----
    UserWarning
        If either ``img1`` or ``img2`` has multiple channels, as SSIM was designed for
        grayscale images.
    UserWarning
        If either ``img1`` or ``img2`` has a value outside of range ``[0, 1]``.
    """
    img_ranges = torch.stack([img1.min(), img1.max(), img2.min(), img2.max()])
    if (img_ranges > 1).any() or (img_ranges < 0).any():
        warnings.warn(
            f"Image range falls outside [0, 1]. {func_name} output may not make sense.",
        )

    if not img1.ndim == img2.ndim == 4:
        raise ValueError(
            "Input images should have four dimensions: (batch, channel, height, width)"
        )
    if img1.shape[-2:] != img2.shape[-2:]:
        raise ValueError("img1 and img2 must have the same height and width!")
    for i in range(2):
        if img1.shape[i] != img2.shape[i] and img1.shape[i] != 1 and img2.shape[i] != 1:
            raise ValueError(
                "Either img1 and img2 should have the same number of "
                "elements in the batch and channel dimensions, or one of "
                "them should be 1! But got shapes "
                f"{img1.shape}, {img2.shape} instead"
            )
    if img1.shape[1] > 1 or img2.shape[1] > 1:
        warnings.warn(
            "SSIM was designed for grayscale images and here it will be"
            " computed separately for each channel (so channels are treated in"
            " the same way as batches).",
        )
    if img1.dtype != img2.dtype:
        raise ValueError("Input images must have same dtype!")

    real_size = min(11, img1.shape[2], img1.shape[3])
    std = torch.as_tensor(1.5).to(img1.device)
    window = circular_gaussian2d(real_size, std=std).to(img1.dtype)

    # these two checks are guaranteed with our above bits, but if we add
    # ability for users to set own window, they'll be necessary
    window_sum = window.sum((-1, -2), keepdim=True)
    if not torch.allclose(window_sum, torch.ones_like(window_sum)):
        warnings.warn("window should have sum of 1! normalizing...")
        window = window / window_sum
    if window.ndim != 4:
        raise ValueError("window must have 4 dimensions!")

    if pad is not False:
        img1 = same_padding(img1, (real_size, real_size), pad_mode=pad)
        img2 = same_padding(img2, (real_size, real_size), pad_mode=pad)

    def windowed_average(img: torch.Tensor) -> torch.Tensor:  # numpydoc ignore=GL08
        padding = 0
        (n_batches, n_channels, _, _) = img.shape
        img = img.reshape(n_batches * n_channels, 1, img.shape[2], img.shape[3])
        img_average = F.conv2d(img, window, padding=padding)
        img_average = img_average.reshape(
            n_batches, n_channels, img_average.shape[2], img_average.shape[3]
        )
        return img_average

    mu1 = windowed_average(img1)
    mu2 = windowed_average(img2)

    mu1_sq = mu1.pow(2)
    mu2_sq = mu2.pow(2)
    mu1_mu2 = mu1 * mu2

    sigma1_sq = windowed_average(img1 * img1) - mu1_sq
    sigma2_sq = windowed_average(img2 * img2) - mu2_sq
    sigma12 = windowed_average(img1 * img2) - mu1_mu2

    C1 = 0.01**2
    C2 = 0.03**2

    # SSIM is the product of a luminance component, a contrast component, and a
    # structure component. The contrast-structure component has to be separated
    # when computing MS-SSIM.
    luminance_map = (2 * mu1_mu2 + C1) / (mu1_sq + mu2_sq + C1)
    contrast_structure_map = (2.0 * sigma12 + C2) / (sigma1_sq + sigma2_sq + C2)
    map_ssim = luminance_map * contrast_structure_map

    # the weight used for stability
    weight = torch.log((1 + sigma1_sq / C2) * (1 + sigma2_sq / C2))
    return map_ssim, contrast_structure_map, weight


def ssim(
    img1: torch.Tensor,
    img2: torch.Tensor,
    weighted: bool = False,
    pad: Literal[False, "constant", "reflect", "replicate", "circular"] = False,
) -> torch.Tensor:
    r"""
    Compute the structural similarity index.

    As described in [1]_, the structural similarity index (SSIM) is a
    perceptual distance metric, giving the distance between two images. SSIM is
    based on three comparison measurements between the two images: luminance,
    contrast, and structure. All of these are computed convolutionally across the
    images. See the references for more information.

    This implementation follows the original implementation, as found at [2]_,
    as well as providing the option to use the weighted version used in [4]_
    (which was shown to consistently improve the image quality prediction on
    the LIVE database). More info can be found at [3]_.

    Note that this is a similarity metric (not a distance), and so 1 means the
    two images are identical and 0 means they're very different. When the two
    images are negatively correlated, SSIM can be negative. SSIM is bounded
    between -1 and 1.

    This function returns the mean SSIM, a scalar-valued metric giving the
    average over the whole image. For the SSIM map (showing the computed value
    across the image), call :func:`ssim_map`.

    Parameters
    ----------
    img1
        The first image or batch of images, of shape (batch, channel, height, width).
    img2
        The second image or batch of images, of shape (batch, channel, height, width).
        The heights and widths of ``img1`` and ``img2`` must be the same. The numbers of
        batches and channels of ``img1`` and ``img2`` need to be broadcastable: either
        they are the same or one of them is 1. The output will be computed separately
        for each channel (so channels are treated in the same way as batches). Both
        images should have values between 0 and 1. Otherwise, the result may be
        inaccurate, and we will raise a warning (but will still compute it).
    weighted
        Whether to use the original, unweighted SSIM version (``False``) as used
        in [1]_ or the weighted version (``True``) as used in [4]_. See Notes
        section for the weight.
    pad :
        If not ``False``, how to pad the image for the convolutions computing the
        local average of each image. See :func:`torch.nn.functional.pad` for how
        these work.

    Returns
    -------
    mssim
        2d tensor of shape (batch, channel) containing the mean SSIM for each
        image, averaged over the whole image.

    Raises
    ------
    ValueError
        If either ``img1`` or ``img2`` is not 4d.
    ValueError
        If ``img1`` and ``img2`` have different height or width.
    ValueError
        If ``img1`` and ``img2`` have different batch or channel, unless one of them has
        a 1 there, so they can be broadcast.
    ValueError
        If ``img1`` and ``img2`` have different dtypes.

    Warns
    -----
    UserWarning
        If either ``img1`` or ``img2`` has multiple channels, as SSIM was designed for
        grayscale images.
    UserWarning
        If at least one scale from either ``img1`` or ``img2`` has height or width of
        less than 11, since SSIM uses an 11x11 convolutional kernel.

    Notes
    -----
    The weight used when ``weighted=True`` is:

    .. math::
       \log((1+\frac{\sigma_1^2}{C_2})(1+\frac{\sigma_2^2}{C_2}))

    where :math:`\sigma_1^2` and :math:`\sigma_2^2` are the variances of ``img1``
    and ``img2``, respectively, and :math:`C_2` is a constant. See [4]_ for more
    details.

    References
    ----------
    .. [1] Z. Wang, A. C. Bovik, H. R. Sheikh, and E. P. Simoncelli, "Image
       quality assessment: From error measurement to structural similarity"
       IEEE Transactions on Image Processing, vol. 13, no. 1, Jan. 2004.
    .. [2] `matlab code <https://www.cns.nyu.edu/~lcv/ssim/ssim_index.m>`_
    .. [3] `project page <https://www.cns.nyu.edu/~lcv/ssim/>`_
    .. [4] Wang, Z., & Simoncelli, E. P. (2008). Maximum differentiation (MAD)
       competition: A methodology for comparing computational models of
       perceptual discriminability. Journal of Vision, 8(12), 1–13.
       https://dx.doi.org/10.1167/8.12.8

    Examples
    --------
    >>> import plenoptic as po
    >>> import torch
    >>> po.tools.set_seed(0)
    >>> img = po.data.einstein()
    >>> po.metric.ssim(img, img + torch.rand_like(img))
    tensor([[0.0519]])
    """
    # these are named map_ssim instead of the perhaps more natural ssim_map
    # because that's the name of a function
    map_ssim, _, weight = _ssim_parts(img1, img2, pad)
    if not weighted:
        mssim = map_ssim.mean((-1, -2))
    else:
        mssim = (map_ssim * weight).sum((-1, -2)) / weight.sum((-1, -2))

    if min(img1.shape[2], img1.shape[3]) < 11:
        warnings.warn(
            "SSIM uses 11x11 convolutional kernel, but the height and/or "
            "the width of the input image is smaller than 11, so the "
            "kernel size is set to be the minimum of these two numbers."
        )
    return mssim


def ssim_map(img1: torch.Tensor, img2: torch.Tensor) -> torch.Tensor:
    """
    Structural similarity index map.

    As described in [1]_, the structural similarity index (SSIM) is a
    perceptual distance metric, giving the distance between two images. SSIM is
    based on three comparison measurements between the two images: luminance,
    contrast, and structure. All of these are computed convolutionally across the
    images. See the references for more information.

    This implementation follows the original implementation, as found at [2]_,
    as well as providing the option to use the weighted version used in [4]_
    (which was shown to consistently improve the image quality prediction on
    the LIVE database). More info can be found at [3]_.

    Note that this is a similarity metric (not a distance), and so 1 means the
    two images are identical and 0 means they're very different. When the two
    images are negatively correlated, SSIM can be negative. SSIM is bounded
    between -1 and 1.

    This function returns the SSIM map, showing the SSIM values across the
    image. For the mean SSIM (a single value metric), call :func:`ssim`.

    Parameters
    ----------
    img1
        The first image or batch of images, of shape (batch, channel, height, width).
    img2
        The second image or batch of images, of shape (batch, channel, height, width).
        The heights and widths of ``img1`` and ``img2`` must be the same. The numbers of
        batches and channels of ``img1`` and ``img2`` need to be broadcastable: either
        they are the same or one of them is 1. The output will be computed separately
        for each channel (so channels are treated in the same way as batches). Both
        images should have values between 0 and 1. Otherwise, the result may be
        inaccurate, and we will raise a warning (but will still compute it).

    Returns
    -------
    ssim_map
        4d tensor containing the map of SSIM values.

    Raises
    ------
    ValueError
        If either ``img1`` or ``img2`` is not 4d.
    ValueError
        If ``img1`` and ``img2`` have different height or width.
    ValueError
        If ``img1`` and ``img2`` have different batch or channel, unless one of them has
        a 1 there, so they can be broadcast.
    ValueError
        If ``img1`` and ``img2`` have different dtypes.

    Warns
    -----
    UserWarning
        If either ``img1`` or ``img2`` has multiple channels, as SSIM was designed for
        grayscale images.
    UserWarning
        If at least one scale from either ``img1`` or ``img2`` has height or width of
        less than 11, since SSIM uses an 11x11 convolutional kernel.

    References
    ----------
    .. [1] Z. Wang, A. C. Bovik, H. R. Sheikh, and E. P. Simoncelli, "Image
       quality assessment: From error measurement to structural similarity"
       IEEE Transactions on Image Processing, vol. 13, no. 1, Jan. 2004.
    .. [2] `matlab code <https://www.cns.nyu.edu/~lcv/ssim/ssim_index.m>`_
    .. [3] `project page <https://www.cns.nyu.edu/~lcv/ssim/>`_
    .. [4] Wang, Z., & Simoncelli, E. P. (2008). Maximum differentiation (MAD)
       competition: A methodology for comparing computational models of
       perceptual discriminability. Journal of Vision, 8(12), 1–13.
       https://dx.doi.org/10.1167/8.12.8

    Examples
    --------
    >>> import plenoptic as po
    >>> import torch
    >>> po.tools.set_seed(0)
    >>> img = po.data.einstein()
    >>> ssim_map = po.metric.ssim_map(img, img + torch.rand_like(img))
    >>> ssim_map.shape
    torch.Size([1, 1, 246, 246])
    """
    if min(img1.shape[2], img1.shape[3]) < 11:
        warnings.warn(
            "SSIM uses 11x11 convolutional kernel, but the height and/or "
            "the width of the input image is smaller than 11, so the "
            "kernel size is set to be the minimum of these two numbers."
        )
    return _ssim_parts(img1, img2)[0]


def ms_ssim(
    img1: torch.Tensor, img2: torch.Tensor, power_factors: torch.Tensor | None = None
) -> torch.Tensor:
    r"""
    Multiscale structural similarity index (MS-SSIM).

    As described in [1]_, multiscale structural similarity index (MS-SSIM) is
    an improvement upon structural similarity index (SSIM) that takes into
    account the perceptual distance between two images on different scales.

    SSIM is based on three comparison measurements between the two images:
    luminance, contrast, and structure. All of these are computed convolutionally
    across the images, producing three maps instead of scalars. The SSIM map is
    the elementwise product of these three maps. See :func:`ssim` and
    :func:`ssim_map` for a full description of SSIM.

    To get images of different scales, average pooling operations with kernel
    size 2 are performed recursively on the input images. The product of
    contrast map and structure map (the "contrast-structure map") is computed
    for all but the coarsest scales, and the overall SSIM map is only computed
    for the coarsest scale. Their mean values are raised to exponents and
    multiplied to produce MS-SSIM:

    .. math::
        MSSSIM = {SSIM}_M^{a_M} \prod_{i=1}^{M-1} ({CS}_i)^{a_i}

    Here :math:`M` is the number of scales, :math:`{CS}_i` is the mean value
    of the contrast-structure map for the i'th finest scale, and :math:`{SSIM}_M`
    is the mean value of the SSIM map for the coarsest scale. If at least one
    of these terms are negative, the value of MS-SSIM is zero. The values of
    :math:`a_i, i=1,...,M` are taken from the argument ``power_factors``.

    Parameters
    ----------
    img1
        The first image or batch of images, of shape (batch, channel, height, width).
    img2
        The second image or batch of images, of shape (batch, channel, height, width).
        The heights and widths of ``img1`` and ``img2`` must be the same. The numbers of
        batches and channels of ``img1`` and ``img2`` need to be broadcastable: either
        they are the same or one of them is 1. The output will be computed separately
        for each channel (so channels are treated in the same way as batches). Both
        images should have values between 0 and 1. Otherwise, the result may be
        inaccurate, and we will raise a warning (but will still compute it).
    power_factors
        Power exponents for the mean values of maps, for different scales (from
        fine to coarse). The length of this array determines the number of scales.
        If ``None``, set to ``[0.0448, 0.2856, 0.3001, 0.2363, 0.1333]``, which is what
        psychophysical experiments in [1]_ found.

    Returns
    -------
    msssim
        2d tensor of shape (batch, channel) containing the MS-SSIM for each image.

    Raises
    ------
    ValueError
        If either ``img1`` or ``img2`` is not 4d.
    ValueError
        If ``img1`` and ``img2`` have different height or width.
    ValueError
        If ``img1`` and ``img2`` have different batch or channel, unless one of them has
        a 1 there, so they can be broadcast.
    ValueError
        If ``img1`` and ``img2`` have different dtypes.

    Warns
    -----
    UserWarning
        If either ``img1`` or ``img2`` has multiple channels, as MS-SSIM was designed
        for grayscale images.
    UserWarning
        If at least one scale from either ``img1`` or ``img2`` has height or width of
        less than 11, since SSIM uses an 11x11 convolutional kernel.

    References
    ----------
    .. [1] Wang, Zhou, Eero P. Simoncelli, and Alan C. Bovik. "Multiscale
       structural similarity for image quality assessment." The Thrity-Seventh
       Asilomar Conference on Signals, Systems & Computers, 2003. Vol. 2. IEEE, 2003.

    Examples
    --------
    >>> import plenoptic as po
    >>> import torch
    >>> po.tools.set_seed(0)
    >>> img = po.data.einstein()
    >>> po.metric.ms_ssim(img, img + torch.rand_like(img))
    tensor([[0.4684]])
    """
    if power_factors is None:
        power_factors = [0.0448, 0.2856, 0.3001, 0.2363, 0.1333]

    def downsample(img: torch.Tensor) -> torch.Tensor:  # numpydoc ignore=GL08
        img = F.pad(img, (0, img.shape[3] % 2, 0, img.shape[2] % 2), mode="replicate")
        img = F.avg_pool2d(img, kernel_size=2)
        return img

    msssim = 1
    for i in range(len(power_factors) - 1):
        _, contrast_structure_map, _ = _ssim_parts(img1, img2, func_name="MS-SSIM")
        msssim *= F.relu(contrast_structure_map.mean((-1, -2))).pow(power_factors[i])
        img1 = downsample(img1)
        img2 = downsample(img2)
    map_ssim, _, _ = _ssim_parts(img1, img2, func_name="MS-SSIM")
    msssim *= F.relu(map_ssim.mean((-1, -2))).pow(power_factors[-1])

    if min(img1.shape[2], img1.shape[3]) < 11:
        warnings.warn(
            "SSIM uses 11x11 convolutional kernel, but for some scales "
            "of the input image, the height and/or the width is smaller "
            "than 11, so the kernel size in SSIM is set to be the "
            "minimum of these two numbers for these scales."
        )
    return msssim


def normalized_laplacian_pyramid(img: torch.Tensor) -> list[torch.Tensor]:
    """
    Compute the normalized Laplacian Pyramid using pre-optimized parameters.

    Model parameters are those used in [1]_, copied from the matlab code used in the
    paper, found at [2]_.

    Parameters
    ----------
    img
        Image, or batch of images of shape (batch, channel, height, width). This
        representation is designed for grayscale images and will be computed separately
        for each channel (so channels are treated in the same way as batches).

    Returns
    -------
    normalized_laplacian_activations
        The normalized Laplacian Pyramid with six scales.

    References
    ----------
    .. [1] Laparra, V., Ballé, J., Berardino, A. and Simoncelli, E.P., 2016. Perceptual
        image quality assessment using a normalized Laplacian pyramid. Electronic
        Imaging, 2016(16), pp.1-6.
    .. [2] `matlab code <https://www.cns.nyu.edu/~lcv/NLPyr/NLP_dist.m>`_

    Examples
    --------
    .. plot::

      >>> import plenoptic as po
      >>> img = po.data.einstein()
      >>> pyramid = po.metric.normalized_laplacian_pyramid(img)
      >>> [p.shape for p in pyramid]
      [torch.Size([1, 1, 256, 256]),
       torch.Size([1, 1, 128, 128]),
       torch.Size([1, 1, 64, 64]),
       torch.Size([1, 1, 32, 32]),
       torch.Size([1, 1, 16, 16]),
       torch.Size([1, 1, 8, 8])]
      >>> po.imshow(pyramid, col_wrap=3)
      <PyrFigure size ...>
    """
    (_, channel, height, width) = img.size()

    N_scales = 6
    spatialpooling_filters = np.load(Path(DIRNAME) / "DN_filts.npy")

    sigmas = np.load(Path(DIRNAME) / "DN_sigmas.npy")

    L = LaplacianPyramid(n_scales=N_scales, scale_filter=True)
    laplacian_activations = L.forward(img)

    padding = 2
    normalized_laplacian_activations = []
    for N_b in range(0, N_scales):
        filt = torch.as_tensor(
            spatialpooling_filters[N_b], dtype=img.dtype, device=img.device
        ).repeat(channel, 1, 1, 1)
        filtered_activations = F.conv2d(
            torch.abs(laplacian_activations[N_b]),
            filt,
            padding=padding,
            groups=channel,
        )
        normalized_laplacian_activations.append(
            laplacian_activations[N_b] / (sigmas[N_b] + filtered_activations)
        )

    return normalized_laplacian_activations


def nlpd(img1: torch.Tensor, img2: torch.Tensor) -> torch.Tensor:
    """
    Compute the normalized Laplacian Pyramid Distance.

    As described in  [1]_, this is an image quality metric based on the transformations
    associated with the early visual system: local luminance subtraction and local
    contrast gain control.

    A laplacian pyramid subtracts a local estimate of the mean luminance at six scales.
    Then a local gain control divides these centered coefficients by a weighted sum of
    absolute values in spatial neighborhood.

    These weights parameters were optimized for redundancy reduction over an training
    database of (undistorted) natural images, as described in the paper. Parameters were
    copied from matlab code used for the paper, found at  [2]_.

    Note that we compute root mean squared error for each scale, and then average over
    these, effectively giving larger weight to the lower frequency coefficients
    (which are fewer in number, due to subsampling).

    Parameters
    ----------
    img1
        The first image or batch of images of shape (batch, channel, height, width).
    img2
        The second image or batch of images of shape (batch, channel, height, width).
        The heights and widths of ``img1`` and ``img2`` must be the same. The numbers of
        batches and channels of ``img1`` and ``img2`` need to be broadcastable: either
        they are the same or one of them is 1. The output will be computed separately
        for each channel (so channels are treated in the same way as batches). Both
        images should have values between 0 and 1. Otherwise, the result may be
        inaccurate, and we will raise a warning (but will still compute it).

    Returns
    -------
    distance
        The normalized Laplacian Pyramid distance, with shape (batch, channel).

    Raises
    ------
    ValueError
        If either ``img1`` or ``img2`` is not 4d.
    ValueError
        If ``img1`` and ``img2`` have different height or width.
    ValueError
        If ``img1`` and ``img2`` have different batch or channel, unless one of them has
        a 1 there, so they can be broadcast.
    ValueError
        If ``img1`` and ``img2`` have different dtypes.

    Warns
    -----
    UserWarning
        If either ``img1`` or ``img2`` has multiple channels, as SSIM was designed for
        grayscale images.
    UserWarning
        If either ``img1`` or ``img2`` has a value outside of range ``[0, 1]``.

    References
    ----------
    .. [1] Laparra, V., Ballé, J., Berardino, A. and Simoncelli, E.P., 2016. Perceptual
        image quality assessment using a normalized Laplacian pyramid. Electronic
        Imaging, 2016(16), pp.1-6.
    .. [2] `matlab code <https://www.cns.nyu.edu/~lcv/NLPyr/NLP_dist.m>`_

    Examples
    --------
    >>> import plenoptic as po
    >>> einstein_img = po.data.einstein()
    >>> curie_img = po.data.curie()
    >>> po.metric.nlpd(einstein_img, curie_img)
    tensor([[1.3507]])
    """
    if not img1.ndim == img2.ndim == 4:
        raise ValueError(
            "Input images should have four dimensions: (batch, channel, height, width)"
        )
    if img1.shape[-2:] != img2.shape[-2:]:
        raise ValueError("img1 and img2 must have the same height and width!")
    for i in range(2):
        if img1.shape[i] != img2.shape[i] and img1.shape[i] != 1 and img2.shape[i] != 1:
            raise ValueError(
                "Either img1 and img2 should have the same number of "
                "elements in the batch and channel dimensions, or one of "
                "them should be 1! But got shapes "
                f"{img1.shape}, {img2.shape} instead"
            )
    if img1.shape[1] > 1 or img2.shape[1] > 1:
        warnings.warn(
            "NLPD was designed for grayscale images and here it will be"
            " computed separately for each channel (so channels are treated in"
            " the same way as batches)."
        )

    img_ranges = torch.stack([img1.min(), img1.max(), img2.min(), img2.max()])
    if (img_ranges > 1).any() or (img_ranges < 0).any():
        warnings.warn(
            "Image range falls outside [0, 1]. NLPD output may not make sense."
        )

    y1 = normalized_laplacian_pyramid(img1)
    y2 = normalized_laplacian_pyramid(img2)

    epsilon = 1e-10  # for optimization purpose (stabilizing the gradient around zero)
    dist = []
    for i in range(6):
        dist.append(torch.sqrt(torch.mean((y1[i] - y2[i]) ** 2, dim=(2, 3)) + epsilon))

    return torch.stack(dist).mean(dim=0)



import nox
import yaml

with open(".pre-commit-config.yaml") as f:
    precommit_config = yaml.safe_load(f)
ruff_version = [r for r in precommit_config["repos"] if "ruff-pre-commit" in r["repo"]]
ruff_version = ruff_version[0]["rev"].replace("v", "")
numpydoc_version = [r for r in precommit_config["repos"] if "numpydoc" in r["repo"]]
numpydoc_version = numpydoc_version[0]["rev"].replace("v", "")


@nox.session
def format(session):
    # run linters
    session.install(f"ruff=={ruff_version}")
    session.run("ruff", "check", "--fix", "--config=pyproject.toml")
    session.run("ruff", "format", "--config=pyproject.toml")


@nox.session
def lint(session):
    # run linters
    session.install(f"ruff=={ruff_version}")
    session.install(f"numpydoc=={numpydoc_version}")
    session.run("ruff", "check", "--config=pyproject.toml")
    session.run("ruff", "format", "--check", "--config=pyproject.toml")
    session.run("numpydoc", "lint", "src")
    session.run("python", "linting/check_docstrings.py", "src")
    session.run("python", "linting/check_sphinx_directives.py", "src")
    session.run("python", "linting/check_tutorials.py", "docs/tutorials/")
    session.run("python", "linting/check_markdown.py", "docs/")
    session.run("python", "linting/ruff_notebooks.py", "docs/tutorials/")
    session.run(
        "python",
        "linting/check_regression_tests.py",
        "tests/test_uploaded_files.py",
        "docs/tutorials/",
    )


@nox.session(name="tests", python=["3.10", "3.11", "3.12"])
def tests(session):
    session.install(".[dev]")
    session.run("pytest")


@nox.session(name="doctests", python=["3.10", "3.11", "3.12"])
def doctests(session):
    session.install(".[dev]")
    session.run("pytest --doctest-modules --doctest-continue-on-failure src/")



"""Example images to use with plenoptic."""  # numpydoc ignore=ES01

from importlib import resources

import torch

from ..tools.data import load_images

__all__ = [
    "einstein",
    "curie",
    "parrot",
    "reptile_skin",
    "color_wheel",
]


def __dir__() -> list[str]:
    return __all__


FILES = resources.files(__name__)


def einstein(as_gray: bool = True) -> torch.Tensor:
    """
    Return an example grayscale image of Albert Einstein.

    Parameters
    ----------
    as_gray :
        Whether to load a single grayscale channel or 3 RGB channels.

    Returns
    -------
    image :
        4d tensor of the image.

    Examples
    --------
    .. plot::

      >>> import plenoptic as po
      >>> einstein = po.data.einstein()
      >>> einstein.shape
      torch.Size([1, 1, 256, 256])
      >>> po.imshow(einstein)
      <PyrFigure size ...>

    >>> import plenoptic as po
    >>> einstein_rgb = po.data.einstein(as_gray=False)
    >>> einstein_rgb.shape
    torch.Size([1, 3, 256, 256])
    """  # numpydoc ignore=ES01
    return load_images(FILES / "einstein.pgm", as_gray=as_gray)


def curie(as_gray: bool = True) -> torch.Tensor:
    """
    Return an example grayscale image of Marie Curie.

    Parameters
    ----------
    as_gray :
        Whether to load a single grayscale channel or 3 RGB channels.

    Returns
    -------
    image :
        4d tensor of the image.

    Examples
    --------
    .. plot::

      >>> import plenoptic as po
      >>> curie = po.data.curie()
      >>> curie.shape
      torch.Size([1, 1, 256, 256])
      >>> po.imshow(curie)
      <PyrFigure size ...>

    >>> import plenoptic as po
    >>> curie_rgb = po.data.curie(as_gray=False)
    >>> curie_rgb.shape
    torch.Size([1, 3, 256, 256])
    """  # numpydoc ignore=ES01
    return load_images(FILES / "curie.pgm", as_gray=as_gray)


def parrot(as_gray: bool = True) -> torch.Tensor:
    """
    Return an example grayscale image of a parrot.

    Parameters
    ----------
    as_gray :
        Whether to load a single grayscale channel or 3 RGB channels.

    Returns
    -------
    image :
        4d tensor of the image.

    Examples
    --------
    .. plot::

      >>> import plenoptic as po
      >>> parrot = po.data.parrot()
      >>> parrot.shape
      torch.Size([1, 1, 254, 266])
      >>> po.imshow(parrot)
      <PyrFigure size ...>

    >>> import plenoptic as po
    >>> parrot_rgb = po.data.parrot(as_gray=False)
    >>> parrot_rgb.shape
    torch.Size([1, 3, 254, 266])
    """  # numpydoc ignore=ES01
    return load_images(FILES / "parrot.png", as_gray=as_gray)


def reptile_skin(as_gray: bool = True) -> torch.Tensor:
    """
    Return an example grayscale image of reptile skin.

    Parameters
    ----------
    as_gray :
        Whether to load a single grayscale channel or 3 RGB channels.

    Returns
    -------
    image :
        4d tensor of the image.

    Examples
    --------
    .. plot::

      >>> import plenoptic as po
      >>> reptile_skin = po.data.reptile_skin()
      >>> reptile_skin.shape
      torch.Size([1, 1, 256, 256])
      >>> po.imshow(reptile_skin)
      <PyrFigure size ...>

    >>> import plenoptic as po
    >>> reptile_skin_rgb = po.data.reptile_skin(as_gray=False)
    >>> reptile_skin_rgb.shape
    torch.Size([1, 3, 256, 256])
    """  # numpydoc ignore=ES01
    return load_images(FILES / "reptile_skin.pgm", as_gray=as_gray)


def color_wheel(as_gray: bool = False) -> torch.Tensor:
    """
    Return an example image of a color wheel.

    Parameters
    ----------
    as_gray :
        Whether to load a single grayscale channel or 3 RGB channels.

    Returns
    -------
    image :
        4d tensor of the image.

    Examples
    --------
    .. plot::

      >>> import plenoptic as po
      >>> color_wheel = po.data.color_wheel()
      >>> color_wheel.shape
      torch.Size([1, 3, 600, 600])
      >>> po.imshow(color_wheel, as_rgb=True, zoom=0.5)
      <PyrFigure size ...>

    .. plot::

      >>> import plenoptic as po
      >>> color_wheel_gray = po.data.color_wheel(as_gray=True)
      >>> color_wheel_gray.shape
      torch.Size([1, 1, 600, 600])
      >>> po.imshow(color_wheel_gray, zoom=0.5)
      <PyrFigure size ...>
    """  # numpydoc ignore=ES01
    return load_images(FILES / "color_wheel.jpg", as_gray=as_gray)



#!/usr/bin/env python3

import ast
import pathlib
import re
import sys

LINK_REGEX = re.escape("](")
# we want basically everything except a backtick
EVERYTHING_BUT_BACKTICK = r"[ A-Za-z_\(\)<>:/\-.=\[\]0-9\\^~{}|'\",]"
BACKTICK_REGEX = rf"(`+{EVERYTHING_BUT_BACKTICK.replace(' ', '')}+?`+_?)"
SPHINX_DIRECTIVE_REGEX = rf":[:a-z]+:(`+{EVERYTHING_BUT_BACKTICK}+?_?`+)"
SPHINX_LINK_REGEX = rf"(`+{EVERYTHING_BUT_BACKTICK}+?`+_)"
# underscores are part of variable names, so we don't want to match them
EVERYTHING_BUT_BACKTICK_ = EVERYTHING_BUT_BACKTICK.replace("_", "")
# None/True/False surrounded by single backtick is also wrong, but will be caught by the
# markdown check.
COMMON_VALS = (
    rf"{EVERYTHING_BUT_BACKTICK_}([Nn]one|[Tt]rue|[Ff]alse){EVERYTHING_BUT_BACKTICK_}"
)
LIBRARIES_XREF = [
    "plenoptic",
    "po",
    "torch",
    "numpy",
    "np",
    "matplotlib",
    "mpl",
    "plt",
    "pyrtools",
    "pt",
]
# we want to grab the references to above libraries that should be xrefs. So we don't
# want to match: proper xrefs (which will have a ~ or ` before the library name) or urls
# (which will have . or / before the name)
XREF_EXCEPT = EVERYTHING_BUT_BACKTICK.replace("~", "").replace(".", "").replace("/", "")
# additionally, = and ( show up in code snippets
XREF_EXCEPT = XREF_EXCEPT.replace("=", "").replace("(", "")


def parse_docstring(docstring):
    docstring = docstring.splitlines()
    docstring = [d.strip() for d in docstring]
    return " ".join(docstring)


def get_docstrings(tree_body):
    docstrings = []
    for node in tree_body:
        try:
            doc = ast.get_docstring(node)
        except TypeError:
            doc = None
        if doc is not None:
            docstrings.append((node.name, parse_docstring(doc)))
    return docstrings


paths = []
for p in sys.argv[1:]:
    p = pathlib.Path(p)
    if p.is_dir():
        p = list(p.glob("**/*.py"))
    elif p.suffix == ".py":
        p = [p]
    else:
        p = []
    paths.extend(p)

links = []
defaults = []
backticks = []
unescaped = []
missing_xref = []
dollarsigns = []

for p in paths:
    with open(p) as f:
        tree = ast.parse(f.read())
    docstrings = get_docstrings(tree.body)

    for name, doc in docstrings:
        if re.findall(LINK_REGEX, doc):
            links.append((p, name))
        if re.findall(r"\$", doc):
            dollarsigns.append((p, name))
        directives = re.findall(SPHINX_DIRECTIVE_REGEX, doc)
        sphinx_link = re.findall(SPHINX_LINK_REGEX, doc)
        # if the word default is preceded / followed by an underscore, then the
        # argument has "default" in its name, so don't match that (using
        # negative look-ahead/behind)
        default = re.findall("(?<!_)[Dd]efault(?!_)", doc)
        if default:
            defaults.append((p, name))
        backtick = [
            x
            for x in re.findall(BACKTICK_REGEX, doc)
            if x not in directives + sphinx_link and not x.startswith("``")
        ]
        if backtick:
            backticks.append((p, name, backtick))
        # See Also and Examples (the last two sections, fortunately) can both contain
        # un-xref'ed references to other libraries or None/False/True
        doc = doc.split("See Also")[0].split("Examples")[0]
        if unesc := re.findall(COMMON_VALS, doc):
            unescaped.append((p, name, unesc))
        for lib in LIBRARIES_XREF:
            # use negative look-behind to avoid stuff in the attribute section
            if xr := re.findall(rf"(?<![a-z]:){XREF_EXCEPT}{lib}\.", doc):
                # the url for torch has `pytorch.` (not `torch.`), which we don't want
                # to match
                if lib == "torch":
                    xr = [x for x in xr if x != "ytorch."]
                if not xr:
                    continue
                missing_xref.append((p, name, xr))
            # we also want to raise an error if we have two backticks before these
            # libraries, which corresponds to monospace (but no link). We *don't* want
            # to match if it looks like an actual call, so we either can have no
            # parentheses or only parentheses with on args in them at the end of the
            # backticks
            if xr := re.findall(rf"``{lib}\.[a-z0-9_\.]+(?:\(\))?``", doc):
                missing_xref.append((p, name, xr))

if backticks or links or unescaped or missing_xref or dollarsigns or defaults:
    if backticks:
        print("The following docstrings appear to contain markdown:")
        for p, name, markup in backticks:
            markup = ", ".join(markup)
            print(f"{p}:{name} {markup}")
        print("\n")
    if links:
        print("The following docstrings appear to contain markdown links:")
        for p, name in links:
            print(f"{p}:{name}")
        print("\n")
    if unescaped:
        print(
            "The following docstrings appear to contain values that should be"
            " surrounded in backticks:"
        )
        for p, name, markup in unescaped:
            print(f"{p}:{name} {markup}")
        print("\n")
    if missing_xref:
        print(
            "The following docstrings appear to contain references to other functions;"
            "should use xref:"
        )
        for p, name, markup in missing_xref:
            print(f"{p}:{name} {markup}")
        print("\n")
    if dollarsigns:
        print(
            "The following docstrings appear to contain latex surrounded by dollar "
            "signs; docstrings should be rst-formatted, use :math:`x` instead of $x$"
        )
        for p, name in dollarsigns:
            print(f"{p}:{name}")
    if defaults:
        print(
            "The following docstrings appear to contain a description of a default "
            "argument. Remove that, it's hard to keep up-to-date."
        )
        for p, name in defaults:
            print(f"{p}:{name}")
    sys.exit(1)



"""
Image quality metrics.

These functions and classes address questions of the form "how different are these
images?"
"""
# ruff: noqa: F401

from .classes import NLP
from .model_metric import model_metric
from .naive import mse
from .perceptual_distance import (
    ms_ssim,
    nlpd,
    normalized_laplacian_pyramid,
    ssim,
    ssim_map,
)



"""
Some useful non-linearities for visual models.

The functions operate on dictionaries or tensors.
"""

import torch

from ...tools import signal
from ...tools.conv import blur_downsample, upsample_blur


def rectangular_to_polar_dict(
    coeff_dict: dict, residuals: bool = False
) -> tuple[dict, dict]:
    """
    Return the complex modulus and the phase of each complex tensor in a dictionary.

    Keys are preserved, with the option of dropping ``"residual_lowpass"`` and
    ``"residual_highpass"`` by setting ``residuals=False``.

    Parameters
    ----------
    coeff_dict
        A dictionary containing complex tensors.
    residuals
        An option to include residuals in the returned ``energy`` dict.

    Returns
    -------
    energy
        The dictionary of :class:`torch.Tensor` containing the local complex
        modulus of ``coeff_dict``.
    state
        The dictionary of :class:`torch.Tensor` containing the local phase of
        ``coeff_dict``.

    See Also
    --------
    :func:`~plenoptic.tools.signal.rectangular_to_polar`
        Same operation on tensors.
    polar_to_rectangular_dict
        The inverse operation.
    local_gain_control_dict
        The analogous function for complex-valued signals.

    Examples
    --------
    .. plot::

        >>> import plenoptic as po
        >>> img = po.data.einstein()
        >>> spyr = po.simul.SteerablePyramidFreq(
        ...     img.shape[-2:], is_complex=True, height=3
        ... )
        >>> coeffs = spyr(img)
        >>> energy, state = po.simul.non_linearities.rectangular_to_polar_dict(coeffs)
        >>> po.pyrshow(energy)
        <PyrFigure size ...>
        >>> po.pyrshow(state)
        <PyrFigure size ...>
    """
    energy = {}
    state = {}
    for key in coeff_dict:
        # ignore residuals
        if isinstance(key, tuple) or not key.startswith("residual"):
            energy[key], state[key] = signal.rectangular_to_polar(coeff_dict[key])

    if residuals:
        energy["residual_lowpass"] = coeff_dict["residual_lowpass"]
        energy["residual_highpass"] = coeff_dict["residual_highpass"]

    return energy, state


def polar_to_rectangular_dict(
    energy: dict,
    state: dict,
) -> dict:
    """
    Return the real and imaginary parts of tensor in a dictionary.

    Keys in the output are identical to those in the input. Will grab residuals from
    ``energy``, if present, with keys ``"residual_highpass"`` and
    ``"residual_lowpass"``.

    Parameters
    ----------
    energy
        The dictionary of :class:`torch.Tensor` containing the local complex
        modulus.
    state
        The dictionary of :class:`torch.Tensor` containing the local phase.

    Returns
    -------
    coeff_dict
       A dictionary containing complex tensors of coefficients.

    See Also
    --------
    :func:`~plenoptic.tools.signal.polar_to_rectangular`
        Same operation on tensors.
    rectangular_to_polar_dict
        The inverse operation.
    local_gain_release_dict
        The analogous function for real-valued signals.

    Examples
    --------
    .. plot::

        >>> import plenoptic as po
        >>> import numpy as np
        >>> import torch
        >>> img = po.data.einstein()
        >>> spyr = po.simul.SteerablePyramidFreq(
        ...     img.shape[-2:], is_complex=True, height=3
        ... )
        >>> coeffs = spyr(img)
        >>> energy, state = po.simul.non_linearities.rectangular_to_polar_dict(
        ...     coeffs, residuals=True
        ... )
        >>> coeffs_back = po.simul.non_linearities.polar_to_rectangular_dict(
        ...     energy, state
        ... )
        >>> all(torch.allclose(coeffs[key], coeffs_back[key]) for key in coeffs)
        True
        >>> po.pyrshow(coeffs_back)
        <PyrFigure size ...>
    """
    coeff_dict = {}
    for key in energy:
        # ignore residuals here
        if isinstance(key, tuple) or not key.startswith("residual"):
            coeff_dict[key] = signal.polar_to_rectangular(energy[key], state[key])

    if "residual_lowpass" in energy:
        coeff_dict["residual_lowpass"] = energy["residual_lowpass"]
        coeff_dict["residual_highpass"] = energy["residual_highpass"]

    return coeff_dict


def local_gain_control(
    x: torch.Tensor, epsilon: float = 1e-8
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Spatially local gain control.

    Compute the local energy and phase of a real-valued tensor.

    Parameters
    ----------
    x
        Tensor of shape (batch, channel, height, width).
    epsilon
        Small constant to avoid division by zero.

    Returns
    -------
    norm
        The local energy of ``x``, shape (batch, channel, height/2,
        width/2).
    direction
        The local phase of ``x`` (a.k.a. local unit vector, or local
        state), shape (batch, channel, height, width).

    See Also
    --------
    local_gain_control_dict
        Same operation on dictionaries.
    local_gain_release
        The inverse operation.
    :func:`~plenoptic.tools.signal.rectangular_to_polar`
        The analogous function for complex-valued signals.

    Notes
    -----
    Norm and direction (analogous to complex modulus and phase) are defined using
    blurring operator and division. Indeed blurring the responses removes high
    frequencies introduced by the squaring operation. In the complex case adding the
    quadrature pair response has the same effect (note that this is most clearly seen in
    the frequency domain). Here computing the direction (phase) reduces to dividing out
    the norm (modulus), indeed the signal only has one real component. This is a
    normalization operation (local unit vector), hence the connection to local gain
    control.

    Examples
    --------
    .. plot::

        >>> import plenoptic as po
        >>> img = po.data.einstein()
        >>> norm, direction = po.simul.non_linearities.local_gain_control(img)
        >>> po.imshow([img, norm, direction], title=["image", "norm", "direction"])
        <PyrFigure size ...>
    """
    # these could be parameters, but no use case so far
    p = 2.0

    norm = blur_downsample(torch.abs(x**p)).pow(1 / p)
    odd = torch.as_tensor(x.shape)[2:4] % 2
    direction = x / (upsample_blur(norm, odd) + epsilon)

    return norm, direction


def local_gain_release(
    norm: torch.Tensor, direction: torch.Tensor, epsilon: float = 1e-8
) -> torch.Tensor:
    """
    Spatially local gain release.

    Convert the local energy and phase to a single real-valued tensor.

    Parameters
    ----------
    norm
        The local energy of a tensor, with shape (batch, channel, height/2,
        width/2).
    direction
        The local phase of a tensor (a.k.a. local unit vector, or local state),
        with shape (batch, channel, height, width).
    epsilon
        Small constant to avoid division by zero.

    Returns
    -------
    x
        Tensor of shape (batch, channel, height, width).

    See Also
    --------
    local_gain_release_dict
        Same operation on dictionaries.
    local_gain_control
        The inverse operation.
    :func:`~plenoptic.tools.signal.polar_to_rectangular`
        The analogous function for complex-valued signals.

    Notes
    -----
    Norm and direction (analogous to complex modulus and phase) are defined using
    blurring operator and division. Indeed blurring the responses removes high
    frequencies introduced by the squaring operation. In the complex case adding the
    quadrature pair response has the same effect (note that this is most clearly seen in
    the frequency domain). Here computing the direction (phase) reduces to dividing out
    the norm (modulus), indeed the signal only has one real component. This is a
    normalization operation (local unit vector), hence the connection to local gain
    control.

    Examples
    --------
    .. plot::

        >>> import plenoptic as po
        >>> img = po.data.einstein()
        >>> norm, direction = po.simul.non_linearities.local_gain_control(img)
        >>> x = po.simul.non_linearities.local_gain_release(norm, direction)
        >>> po.imshow(
        ...     [img, x, img - x],
        ...     title=["Original image", "Gain release output", "Difference"],
        ... )
        <PyrFigure size ...>
    """
    odd = torch.as_tensor(direction.shape)[2:4] % 2
    x = direction * (upsample_blur(norm, odd) + epsilon)
    return x


def local_gain_control_dict(
    coeff_dict: dict, residuals: bool = True
) -> tuple[dict, dict]:
    """
    Spatially local gain control, for each element in a dictionary.

    For more details, see :func:`local_gain_control`.

    Parameters
    ----------
    coeff_dict
        A dictionary containing tensors of shape (batch, channel, height, width).
    residuals
        An option to carry around residuals in the energy dict.
        Note that the transformation is not applied to the residuals,
        that is dictionary elements with a key starting in "residual".

    Returns
    -------
    energy
        The dictionary of :class:`torch.Tensor` containing the local energy of
        ``x``.
    state
        The dictionary of :class:`torch.Tensor` containing the local phase of
        ``x``.

    See Also
    --------
    local_gain_control
        Same operation on tensors.
    local_gain_release_dict
        The inverse operation.
    rectangular_to_polar_dict
        The analogous function for complex-valued signals.

    Notes
    -----
    Note that energy and state are not computed on the residuals.

    Examples
    --------
    .. plot::

        >>> import plenoptic as po
        >>> img = po.data.einstein()
        >>> spyr = po.simul.SteerablePyramidFreq(img.shape[-2:], height=3)
        >>> coeffs = spyr(img)
        >>> energy, state = po.simul.non_linearities.local_gain_control_dict(coeffs)
        >>> po.pyrshow(energy)
        <PyrFigure size ...>
        >>> po.pyrshow(state)
        <PyrFigure size ...>
    """
    energy = {}
    state = {}

    for key in coeff_dict:
        if isinstance(key, tuple) or not key.startswith("residual"):
            energy[key], state[key] = local_gain_control(coeff_dict[key])

    if residuals:
        energy["residual_lowpass"] = coeff_dict["residual_lowpass"]
        energy["residual_highpass"] = coeff_dict["residual_highpass"]

    return energy, state


def local_gain_release_dict(energy: dict, state: dict, residuals: bool = True) -> dict:
    """
    Spatially local gain release, for each element in a dictionary.

    For more details, see :func:`local_gain_release`.

    Parameters
    ----------
    energy
        The dictionary of :class:`torch.Tensor` containing the local energy of
        ``x``.
    state
        The dictionary of :class:`torch.Tensor` containing the local phase of
        ``x``.
    residuals
        An option to carry around residuals in the energy dict.
        Note that the transformation is not applied to the residuals,
        that is dictionary elements with a key starting in "residual".

    Returns
    -------
    coeff_dict
        A dictionary containing tensors of shape (batch, channel, height, width).

    See Also
    --------
    local_gain_release
        Same operation on tensors.
    local_gain_control_dict
        The inverse operation.
    polar_to_rectangular_dict
        The analogous function for complex-valued signals.

    Examples
    --------
    .. plot::

        >>> import plenoptic as po
        >>> import torch
        >>> img = po.data.einstein()
        >>> spyr = po.simul.SteerablePyramidFreq(img.shape[-2:], height=3)
        >>> coeffs = spyr(img)
        >>> energy, state = po.simul.non_linearities.local_gain_control_dict(coeffs)
        >>> coeffs_dict = po.simul.non_linearities.local_gain_release_dict(
        ...     energy, state
        ... )
        >>> all([torch.allclose(coeffs[k], coeffs_dict[k]) for k in coeffs.keys()])
        True
        >>> po.pyrshow(coeffs_dict)
        <PyrFigure size ...>
    """
    coeff_dict = {}

    for key in energy:
        if isinstance(key, tuple) or not key.startswith("residual"):
            coeff_dict[key] = local_gain_release(energy[key], state[key])

    if residuals:
        coeff_dict["residual_lowpass"] = energy["residual_lowpass"]
        coeff_dict["residual_highpass"] = energy["residual_highpass"]

    return coeff_dict



"""Canonical computations.

These functions and classes may be useful when building models.
"""
# ignore F401 (unused import) and F403 (from module import *)
# ruff: noqa: F401, F403

from .filters import *
from .laplacian_pyramid import LaplacianPyramid
from .non_linearities import *
from .steerable_pyramid_freq import SteerablePyramidFreq



from contextlib import nullcontext as does_not_raise

import einops
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pyrtools as pt
import pytest
import torch

import plenoptic as po
from conftest import DEVICE, IMG_DIR

# use the html backend, so we don't need to have ffmpeg
mpl.rcParams["animation.writer"] = "html"
# necessary to avoid issues with animate:
# https://github.com/matplotlib/matplotlib/issues/10287/
mpl.use("agg")


class TestDisplay:
    def test_update_plot_line(self):
        x = np.linspace(0, 100)
        y1 = np.random.rand(*x.shape)
        y2 = np.random.rand(*x.shape)
        fig, ax = plt.subplots(1, 1)
        ax.plot(x, y1, "-o", label="hi")
        po.tools.update_plot(
            ax, torch.as_tensor(y2, device=DEVICE).reshape(1, 1, len(x))
        )
        assert len(ax.lines) == 1, "Too many lines were plotted!"
        _, ax_y = ax.lines[0].get_data()
        if not np.allclose(ax_y, y2):
            raise Exception("Didn't update line correctly!")
        plt.close(fig)

    @pytest.mark.parametrize("how", ["dict", "tensor"])
    def test_update_plot_line_multi_axes(self, how):
        x = np.linspace(0, 100)
        y1 = np.random.rand(*x.shape)
        y2 = np.random.rand(2, *y1.shape)
        if how == "tensor":
            y2 = torch.as_tensor(y2, device=DEVICE).reshape(1, 2, *y1.shape)
        elif how == "dict":
            y2 = {
                i: torch.as_tensor(y2[i], device=DEVICE).reshape(1, 1, *y1.shape)
                for i in range(2)
            }
        fig, axes = plt.subplots(1, 2)
        for ax in axes:
            ax.plot(x, y1, "-o", label="hi")
        po.tools.update_plot(axes, y2)
        for i, ax in enumerate(axes):
            assert len(ax.lines) == 1, "Too many lines were plotted!"
            _, ax_y = ax.lines[0].get_data()
            y_check = y2[0, i] if how == "tensor" else y2[i]

            if not np.allclose(ax_y, po.to_numpy(y_check)):
                raise Exception("Didn't update line correctly!")
        plt.close(fig)

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("figsize", [None, (5, 5), (5.0, 5.0), (10, 5)])
    @pytest.mark.parametrize("ax", [False, True])
    def test_plot_representation_figsize(self, model, figsize, ax, einstein_img):
        expectation = does_not_raise()
        if ax:
            fig, ax = plt.subplots(1, 1)
            if figsize is not None:
                expectation = pytest.raises(ValueError, match="figsize can't be set")
        else:
            ax = None
        with expectation:
            axes = po.tools.plot_representation(
                data=model(einstein_img), ax=ax, figsize=figsize
            )
            fig = axes[0].figure
        plt.close(fig)

    @pytest.mark.parametrize("how", ["dict-single", "dict-multi", "tensor"])
    def test_update_plot_line_multi_channel(self, how):
        n_data = 1 if how == "dict-single" else 2

        x = np.linspace(0, 100)
        y1 = np.random.rand(2, *x.shape)
        y2 = np.random.rand(n_data, *x.shape)
        if how == "tensor":
            y2 = torch.as_tensor(y2, device=DEVICE).reshape(1, 2, len(x))
        elif how == "dict-multi":
            y2 = {
                i: torch.as_tensor(y2[i], device=DEVICE).reshape(1, 1, len(x))
                for i in range(2)
            }
        elif how == "dict-single":
            y2 = {0: torch.as_tensor(y2[0], device=DEVICE).reshape(1, 1, len(x))}
        fig, ax = plt.subplots(1, 1)
        for i in range(2):
            ax.plot(x, y1[i], label=i)
        po.tools.update_plot(ax, y2)
        assert len(ax.lines) == 2, "Incorrect number of lines were plotted!"
        for i in range(2):
            _, ax_y = ax.lines[i].get_data()
            if how == "tensor":
                y_check = y2[0, i]
            elif how == "dict-multi":
                y_check = y2[i]
            elif how == "dict-single":
                y_check = {0: y2[0], 1: y1[1]}[i]
            if not np.allclose(ax_y, po.to_numpy(y_check)):
                raise Exception("Didn't update line correctly!")
        plt.close(fig)

    def test_update_plot_stem(self):
        x = np.linspace(0, 100)
        y1 = np.random.rand(*x.shape)
        y2 = np.random.rand(*x.shape)
        fig, ax = plt.subplots(1, 1)
        ax.stem(x, y1, "-o", label="hi")
        po.tools.update_plot(
            ax, torch.as_tensor(y2, device=DEVICE).reshape(1, 1, len(x))
        )
        assert len(ax.containers) == 1, "Too many stems were plotted!"
        ax_y = ax.containers[0].markerline.get_ydata()
        if not np.allclose(ax_y, y2):
            raise Exception("Didn't update stems correctly!")
        plt.close(fig)

    @pytest.mark.parametrize("how", ["dict", "tensor"])
    def test_update_plot_stem_multi_axes(self, how):
        x = np.linspace(0, 100)
        y1 = np.random.rand(*x.shape)
        y2 = np.random.rand(2, *y1.shape)
        if how == "tensor":
            y2 = torch.as_tensor(y2, device=DEVICE).reshape(1, 2, *y1.shape)
        elif how == "dict":
            y2 = {
                i: torch.as_tensor(y2[i], device=DEVICE).reshape(1, 1, *y1.shape)
                for i in range(2)
            }
        fig, axes = plt.subplots(1, 2)
        for ax in axes:
            ax.stem(x, y1, label="hi")
        po.tools.update_plot(axes, y2)
        for i, ax in enumerate(axes):
            assert len(ax.containers) == 1, "Too many stems were plotted!"
            ax_y = ax.containers[0].markerline.get_ydata()
            y_check = y2[0, i] if how == "tensor" else y2[i]

            if not np.allclose(ax_y, po.to_numpy(y_check)):
                raise Exception("Didn't update stem correctly!")
        plt.close(fig)

    @pytest.mark.parametrize("how", ["dict-single", "dict-multi", "tensor"])
    def test_update_plot_stem_multi_channel(self, how):
        n_data = 1 if how == "dict-single" else 2

        x = np.linspace(0, 100)
        y1 = np.random.rand(2, *x.shape)
        y2 = np.random.rand(n_data, *x.shape)
        if how == "tensor":
            y2 = torch.as_tensor(y2, device=DEVICE).reshape(1, 2, len(x))
        elif how == "dict-multi":
            y2 = {
                i: torch.as_tensor(y2[i], device=DEVICE).reshape(1, 1, len(x))
                for i in range(2)
            }
        elif how == "dict-single":
            y2 = {0: torch.as_tensor(y2[0], device=DEVICE).reshape(1, 1, len(x))}
        fig, ax = plt.subplots(1, 1)
        for i in range(2):
            ax.stem(x, y1[i], label=str(i))
        po.tools.update_plot(ax, y2)
        assert len(ax.containers) == 2, "Incorrect number of stems were plotted!"
        for i in range(2):
            ax_y = ax.containers[i].markerline.get_ydata()
            if how == "tensor":
                y_check = y2[0, i]
            elif how == "dict-multi":
                y_check = y2[i]
            elif how == "dict-single":
                y_check = {0: y2[0], 1: y1[1]}[i]
            if not np.allclose(ax_y, po.to_numpy(y_check)):
                raise Exception("Didn't update stem correctly!")
        plt.close(fig)

    def test_update_plot_image(self):
        y1 = np.random.rand(1, 1, 100, 100)
        y2 = np.random.rand(*y1.shape)
        fig = pt.imshow(y1.squeeze())
        ax = fig.axes[0]
        po.tools.update_plot(ax, torch.as_tensor(y2, device=DEVICE))
        assert len(ax.images) == 1, "Too many images were plotted!"
        ax_y = ax.images[0].get_array().data
        if not np.allclose(ax_y, y2):
            raise Exception("Didn't update image correctly!")
        plt.close(fig)

    @pytest.mark.parametrize("how", ["dict", "tensor"])
    def test_update_plot_image_multi_axes(self, how):
        y1 = np.random.rand(1, 2, 100, 100)
        y2 = np.random.rand(1, 2, 100, 100)
        if how == "tensor":
            y2 = torch.as_tensor(y2, device=DEVICE)
        elif how == "dict":
            y2 = {
                i: torch.as_tensor(y2[0, i], device=DEVICE).reshape(1, 1, 100, 100)
                for i in range(2)
            }
        fig = pt.imshow([y for y in y1.squeeze()])
        po.tools.update_plot(fig.axes, y2)
        for i, ax in enumerate(fig.axes):
            assert len(ax.images) == 1, "Too many lines were plotted!"
            ax_y = ax.images[0].get_array().data

            y_check = y2[0, i] if how == "tensor" else y2[i]

            if not np.allclose(ax_y, po.to_numpy(y_check)):
                raise Exception("Didn't update image correctly!")
        plt.close(fig)

    def test_update_plot_scatter(self):
        x1 = np.random.rand(100)
        x2 = np.random.rand(100)
        y1 = np.random.rand(*x1.shape)
        y2 = np.random.rand(*x2.shape)
        fig, ax = plt.subplots(1, 1)
        ax.scatter(x1, y1)
        data = torch.stack(
            (torch.as_tensor(x2, device=DEVICE), torch.as_tensor(y2, device=DEVICE)), -1
        ).reshape(1, 1, len(x2), 2)
        po.tools.update_plot(ax, data)
        assert len(ax.collections) == 1, "Too many scatter plots created"
        ax_data = ax.collections[0].get_offsets()
        if not np.allclose(ax_data, po.to_numpy(data)):
            raise Exception("Didn't update points of the scatter plot correctly!")
        plt.close(fig)

    @pytest.mark.parametrize("how", ["dict", "tensor"])
    def test_update_plot_scatter_multi_axes(self, how):
        x1 = np.random.rand(100)
        x2 = np.random.rand(2, 100)
        y1 = np.random.rand(*x1.shape)
        y2 = np.random.rand(2, *y1.shape)
        if how == "tensor":
            data = torch.stack(
                (
                    torch.as_tensor(x2, device=DEVICE),
                    torch.as_tensor(y2, device=DEVICE),
                ),
                -1,
            ).reshape(1, 2, len(x1), 2)
        elif how == "dict":
            data = {
                i: torch.stack(
                    (
                        torch.as_tensor(x2[i], device=DEVICE),
                        torch.as_tensor(y2[i], device=DEVICE),
                    ),
                    -1,
                ).reshape(1, 1, len(x1), 2)
                for i in range(2)
            }
        fig, axes = plt.subplots(1, 2)
        for ax in axes:
            ax.scatter(x1, y1)
        po.tools.update_plot(axes, data)
        for i, ax in enumerate(axes):
            assert len(ax.collections) == 1, "Too many scatter plots created"
            ax_data = ax.collections[0].get_offsets()
            data_check = data[0, i] if how == "tensor" else data[i]

            if not np.allclose(ax_data, po.to_numpy(data_check)):
                raise Exception("Didn't update points of the scatter plot correctly!")
        plt.close(fig)

    @pytest.mark.parametrize("how", ["dict-single", "dict-multi", "tensor"])
    def test_update_plot_scatter_multi_channel(self, how):
        n_data = 1 if how == "dict-single" else 2
        x1 = np.random.rand(2, 100)
        x2 = np.random.rand(n_data, 100)
        y1 = np.random.rand(*x1.shape)
        y2 = np.random.rand(*x2.shape)
        if how == "tensor":
            data = torch.stack(
                (
                    torch.as_tensor(x2, device=DEVICE),
                    torch.as_tensor(y2, device=DEVICE),
                ),
                -1,
            ).reshape(1, 2, x1.shape[-1], 2)
        elif how == "dict-multi":
            data = {
                i: torch.stack(
                    (
                        torch.as_tensor(x2[i], device=DEVICE),
                        torch.as_tensor(y2[i], device=DEVICE),
                    ),
                    -1,
                ).reshape(1, 1, x1.shape[-1], 2)
                for i in range(2)
            }
        elif how == "dict-single":
            data = {
                0: torch.stack(
                    (
                        torch.as_tensor(x2[0], device=DEVICE),
                        torch.as_tensor(y2[0], device=DEVICE),
                    ),
                    -1,
                ).reshape(1, 1, x1.shape[-1], 2)
            }
        fig, ax = plt.subplots(1, 1)
        for i in range(2):
            ax.scatter(x1[i], y1[i], label=i)
        po.tools.update_plot(ax, data)
        assert len(ax.collections) == 2, "Incorrect number of scatter plots created"
        for i in range(2):
            ax_data = ax.collections[i].get_offsets()
            if how == "tensor":
                data_check = data[0, i]
            elif how == "dict-multi":
                data_check = data[i]
            elif how == "dict-single":
                tmp = torch.stack(
                    (
                        torch.as_tensor(x1, device=DEVICE),
                        torch.as_tensor(y1, device=DEVICE),
                    ),
                    -1,
                )
                data_check = {0: data[0], 1: tmp[1]}[i]
            if not np.allclose(ax_data, po.to_numpy(data_check)):
                raise Exception("Didn't update points of the scatter plot correctly!")

    def test_update_plot_mixed_multi_axes(self):
        x1 = np.linspace(0, 1, 100)
        x2 = np.random.rand(2, 100)
        y1 = np.random.rand(*x1.shape)
        y2 = np.random.rand(*x2.shape)
        data = {
            0: torch.stack(
                (
                    torch.as_tensor(x2[0], device=DEVICE),
                    torch.as_tensor(y2[0], device=DEVICE),
                ),
                -1,
            ).reshape(1, 1, x2.shape[-1], 2)
        }
        data[1] = torch.as_tensor(y2[1], device=DEVICE).reshape(1, 1, x2.shape[-1])
        fig, axes = plt.subplots(1, 2)
        for i, ax in enumerate(axes):
            if i == 0:
                ax.scatter(x1, y1)
            else:
                ax.plot(x1, y1)
        po.tools.update_plot(axes, data)
        for i, ax in enumerate(axes):
            if i == 0:
                assert len(ax.collections) == 1, "Too many scatter plots created"
                assert len(ax.lines) == 0, "Too many lines created"
                ax_data = ax.collections[0].get_offsets()
            else:
                assert len(ax.collections) == 0, "Too many scatter plots created"
                assert len(ax.lines) == 1, "Too many lines created"
                _, ax_data = ax.lines[0].get_data()
            if not np.allclose(ax_data, po.to_numpy(data[i])):
                raise Exception("Didn't update points of the scatter plot correctly!")
        plt.close(fig)

    @pytest.mark.parametrize("as_rgb", [True, False])
    @pytest.mark.parametrize("channel_idx", [None, 0, [0, 1]])
    @pytest.mark.parametrize("batch_idx", [None, 0, [0, 1]])
    @pytest.mark.parametrize("is_complex", [False, "logpolar", "rectangular", "polar"])
    @pytest.mark.parametrize("mini_im", [True, False])
    # test the edge cases where we try to plot a tensor that's (b, c, 1, w) or
    # (b, c, h, 1)
    @pytest.mark.parametrize("one_dim", [False, "h", "w"])
    def test_imshow(self, as_rgb, channel_idx, batch_idx, is_complex, mini_im, one_dim):
        expectation = does_not_raise()
        if one_dim == "h":
            im_shape = [2, 4, 1, 5]
        elif one_dim == "w":
            im_shape = [2, 4, 5, 1]
        else:
            im_shape = [2, 4, 5, 5]
        if is_complex:
            dtype = torch.complex64
            # this is 4 (the four channels) * 2 (the two batches) * 2 (the two
            # complex components)
            n_axes = 16
        else:
            dtype = torch.float32
            # this is 4 (the four channels) * 2 (the two batches)
            n_axes = 8
        im = torch.rand(im_shape, dtype=dtype)
        if mini_im:
            # n_axes here follows the same logic as above
            if is_complex:
                n_axes += 16
            else:
                n_axes += 8
            # same number of batches and channels, then double the height and
            # width
            shape = im_shape[:2] + [i * 2 for i in im_shape[-2:]]
            im = [im, torch.rand(shape, dtype=dtype)]
        if not is_complex:
            # need to change this to one of the acceptable strings
            is_complex = "rectangular"
        if batch_idx is None and channel_idx is None and not as_rgb:
            # then we'd have a 4d array we want to plot in grayscale -- don't
            # know how to do that
            expectation = pytest.raises(
                ValueError,
                match=(
                    "Don't know how to plot non-rgb images with more than one "
                    "channel and batch"
                ),
            )
        else:
            if batch_idx is not None:
                # then we're only plotting one of the two batches
                n_axes /= 2
            if channel_idx is not None:
                # then we're only plotting one of the four channels
                n_axes /= 4
                # if channel_idx is not None, then we don't have all the
                # channels necessary for plotting RGB, so this will fail
                if as_rgb:
                    expectation = pytest.raises(
                        ValueError, match="If as_rgb is True, then channel"
                    )
            # when channel_idx=0, as_rgb does nothing, so don't want to
            # double-count
            elif as_rgb:
                # if we're plotting as_rgb, the four channels just specify
                # RGBA, so we only have one image for them
                n_axes /= 4
        if isinstance(batch_idx, list) or isinstance(channel_idx, list):
            # neither of these are supported
            expectation = pytest.raises(
                TypeError, match=r".* must be an int or None but got"
            )
        with expectation:
            fig = po.imshow(
                im,
                as_rgb=as_rgb,
                channel_idx=channel_idx,
                batch_idx=batch_idx,
                plot_complex=is_complex,
            )
            assert len(fig.axes) == n_axes, (
                f"Created {len(fig.axes)} axes, but expected {n_axes}!"
                " Probably plotting color as grayscale or vice versa"
            )
            plt.close(fig)

    @pytest.mark.parametrize("channel_idx", [None, 0, [0, 1]])
    @pytest.mark.parametrize("batch_idx", [None, 0, [0, 1]])
    @pytest.mark.parametrize("show_residuals", [True, False])
    @pytest.mark.parametrize(
        "plot_complex", [False, "rectangular", "polar", "logpolar"]
    )
    def test_pyrshow(
        self, channel_idx, batch_idx, show_residuals, plot_complex, curie_img
    ):
        if not plot_complex:
            is_complex = False
            # the default value to plot_complex
            plot_complex = "rectangular"
        else:
            is_complex = True
        steerpyr = po.simul.SteerablePyramidFreq(
            (32, 32), height=2, order=1, is_complex=is_complex
        ).to(DEVICE)
        expectation = does_not_raise()
        if not isinstance(channel_idx, int) or not isinstance(batch_idx, int):
            expectation = pytest.raises(TypeError, match=r".* must be an int but got")
        n_axes = 4
        if steerpyr.is_complex:
            n_axes *= 2
        if show_residuals:
            n_axes += 2
        img = curie_img.clone()
        img = img[..., : steerpyr.lo0mask.shape[-2], : steerpyr.lo0mask.shape[-1]]
        coeffs = steerpyr(img)
        with expectation:
            fig = po.pyrshow(
                coeffs,
                show_residuals=show_residuals,
                plot_complex=plot_complex,
                batch_idx=batch_idx,
                channel_idx=channel_idx,
            )
            # get all the axes that have an image (so, get all non-empty
            # axes)
            axes = [ax for ax in fig.axes if ax.images]
            if len(axes) != n_axes:
                raise Exception(f"Created {len(fig.axes)} axes, but expected {n_axes}!")
            plt.close(fig)

    def test_display_test_signals(self, basic_stim):
        po.imshow(basic_stim)

    @pytest.mark.parametrize("as_rgb", [True, False])
    @pytest.mark.parametrize("channel_idx", [None, 0, [0, 1]])
    @pytest.mark.parametrize("batch_idx", [None, 0, [0, 1]])
    @pytest.mark.parametrize("is_complex", [False, "logpolar", "rectangular", "polar"])
    @pytest.mark.parametrize("mini_vid", [True, False])
    def test_animshow(self, as_rgb, channel_idx, batch_idx, is_complex, mini_vid):
        expectation = does_not_raise()
        if is_complex:
            # this is 2 (the two complex components) * 4 (the four channels) *
            # 2 (the two batches)
            n_axes = 16
            dtype = torch.complex64
        else:
            # this is 4 (the four channels) * 2 (the two batches)
            n_axes = 8
            dtype = torch.float32
        vid = torch.rand((2, 4, 10, 10, 10), dtype=dtype)
        if mini_vid:
            # n_axes here follows the same logic as above
            if is_complex:
                n_axes += 16
            else:
                n_axes += 8
            # same number of batches, channels, and frames, then half the
            # height and width
            shape = [2, 4, 10, 5, 5]
            vid = [vid, torch.rand(shape, dtype=dtype)]
        if not is_complex:
            # need to change this to one of the acceptable strings
            is_complex = "rectangular"
        if batch_idx is None and channel_idx is None and not as_rgb:
            # then we'd have a 4d array we want to plot in grayscale -- don't
            # know how to do that
            expectation = pytest.raises(
                ValueError,
                match=(
                    "Don't know how to plot non-rgb images with more than one "
                    "channel and batch"
                ),
            )
        else:
            if batch_idx is not None:
                # then we're only plotting one of the two batches
                n_axes /= 2
            if channel_idx is not None:
                # then we're only plotting one of the four channels
                n_axes /= 4
                # if channel_idx is not None, then we don't have all the
                # channels necessary for plotting RGB, so this will fail
                if as_rgb:
                    expectation = pytest.raises(
                        ValueError, match="If as_rgb is True, then channel"
                    )
            # when channel_idx=0, as_rgb does nothing, so don't want to
            # double-count
            elif as_rgb:
                # if we're plotting as_rgb, the four channels just specify
                # RGBA, so we only have one video for them
                n_axes /= 4
        if isinstance(batch_idx, list) or isinstance(channel_idx, list):
            # neither of these are supported
            expectation = pytest.raises(
                TypeError, match="must be an int or None but got"
            )
        with expectation:
            anim = po.animshow(
                vid,
                as_rgb=as_rgb,
                channel_idx=channel_idx,
                batch_idx=batch_idx,
                plot_complex=is_complex,
            )
            fig = anim._fig
            assert len(fig.axes) == n_axes, (
                f"Created {len(fig.axes)} axes, but expected {n_axes}!"
                " Probably plotting color as grayscale or vice versa"
            )
            plt.close(fig)

    def test_update_plot_shape_fail(self, einstein_img):
        # update_plot expects 3 or 4d data -- this checks that update_plot
        # fails with 2d data and raises the proper exception
        fig = po.imshow(einstein_img)
        with pytest.raises(ValueError, match="3 or 4 dimensional"):
            po.tools.update_plot(fig.axes[0], einstein_img.squeeze())

    def test_synthesis_plot_shape_fail(self, einstein_img):
        # Synthesis plot_synthesis_status and animate expect 3 or 4d data --
        # this checks that plot_synthesis_status() and animate() both fail with
        # 2d data and raise the proper exception
        class TestModel(po.simul.LinearNonlinear):
            def __init__(self):
                super().__init__((7, 7))

            def forward(self, *args, **kwargs):
                return super().forward(*args, **kwargs)

        model = TestModel().to(DEVICE)
        model.eval()
        po.tools.remove_grad(model)
        met = po.synth.Metamer(einstein_img, model)
        met.synthesize(max_iter=3, store_progress=True)
        met._metamer = met.metamer.squeeze()
        with pytest.raises(ValueError, match="3 or 4d"):
            po.synth.metamer.plot_synthesis_status(met)
        with pytest.raises(ValueError, match="3 or 4d"):
            po.synth.metamer.animate(met)

    @pytest.mark.parametrize("zoom", [None, 0.5, 1, 3, 0, -1, 1.5, 1.1])
    @pytest.mark.parametrize("func", ["imshow", "animshow"])
    def test_zoom(self, zoom, func):
        fig, ax = plt.subplots(1, 1, figsize=(5, 5))
        if func == "imshow":
            x = torch.rand(1, 1, 32, 32)
            func = po.imshow
        elif func == "animshow":
            x = torch.rand(1, 1, 10, 32, 32)
            func = po.animshow
        if zoom is not None and zoom <= 0:
            expectation = pytest.raises(ValueError, match="zoom must be positive")
        elif zoom is not None and int(zoom * 32) != zoom * 32:
            expectation = pytest.raises(
                Exception, match=r"zoom \* signal.shape must result in integers"
            )
        else:
            expectation = does_not_raise()
        with expectation:
            func(x, ax=ax, zoom=zoom)
            if zoom is None:
                zoom = 12
            zoom_title = float(ax.get_title().split("*")[1])
            if zoom != zoom_title:
                raise ValueError(
                    f"Zoom didn't work as expected, expected {zoom} but"
                    f" got {zoom_title}!"
                )
            plt.close(fig)


def template_test_synthesis_all_plot(
    synthesis_object,
    iteration,
    display_synth,
    loss,
    representation_error,
    pixel_values,
    fig_creation,
):
    # template function to test whether we can plot all possible combinations
    # of plots. test_custom_fig tests whether these animate correctly. Any
    # synthesis object that has had synthesis() called should work with this
    if sum([display_synth, loss, representation_error, pixel_values]) == 0:
        # then there's nothing to plot here
        return
    included_plots = []
    plot_kwargs = {}
    # need to figure out which plotting function to call
    if isinstance(synthesis_object, po.synth.Metamer):
        containing_file = po.synth.metamer
        as_rgb = synthesis_object.image.shape[1] > 1
        plot_kwargs["plot_representation_error_as_rgb"] = as_rgb
        if display_synth:
            included_plots.append("display_metamer")
    elif isinstance(synthesis_object, po.synth.MADCompetition):
        containing_file = po.synth.mad_competition
        if display_synth:
            included_plots.append("display_mad_image")
    if loss:
        included_plots.append("plot_loss")
    if representation_error:
        included_plots.append("plot_representation_error")
    if pixel_values:
        included_plots.append("plot_pixel_values")
    width_ratios = {}
    if fig_creation.startswith("auto"):
        fig = None
        axes_idx = {}
        if fig_creation.endswith("ratios"):
            if loss:
                width_ratios["plot_loss"] = 2
            elif display_synth:
                if isinstance(synthesis_object, po.synth.Metamer):
                    width_ratios["display_metamer"] = 2
                elif isinstance(synthesis_object, po.synth.MADCompetition):
                    width_ratios["display_mad_image"] = 2
    elif fig_creation.startswith("pass"):
        fig, axes, axes_idx = containing_file._setup_synthesis_fig(
            None, {}, None, included_plots=included_plots
        )
        if fig_creation.endswith("without"):
            axes_idx = {}
    fig, _ = containing_file.plot_synthesis_status(
        synthesis_object,
        iteration=iteration,
        included_plots=included_plots,
        fig=fig,
        axes_idx=axes_idx,
        **plot_kwargs,
        width_ratios=width_ratios,
    )
    plt.close(fig)


def template_test_synthesis_custom_fig(synthesis_object, func, fig_creation, tmp_path):
    # template function to test whether we can create our own figure and pass
    # it to the plotting and animating functions, specifying some or all of the
    # locations for the plots. Any synthesis object that has had synthesis()
    # called should work with this
    plot_kwargs = {}
    included_plots = ["plot_loss", "plot_pixel_values"]
    # need to figure out which plotting function to call
    if isinstance(synthesis_object, po.synth.Metamer):
        containing_file = po.synth.metamer
        as_rgb = synthesis_object.image.shape[1] > 1
        plot_kwargs["plot_representation_error_as_rgb"] = as_rgb
        included_plots.append("plot_representation_error")
        if fig_creation.endswith("extra"):
            included_plots.append("display_metamer")
        axes_idx = {"display_metamer": 0, "plot_representation_error": 8}
    elif isinstance(synthesis_object, po.synth.MADCompetition):
        containing_file = po.synth.mad_competition
        if fig_creation.endswith("extra"):
            included_plots.append("display_mad_image")
        axes_idx = {"display_mad_image": 0}
    fig, axes = plt.subplots(3, 3, figsize=(35, 17))
    if "-" in fig_creation:
        axes_idx["misc"] = [1, 4]
    if fig_creation.split("-")[-1] not in ["without"]:
        axes_idx.update({"plot_loss": 6, "plot_pixel_values": 7})
    if func == "plot" or fig_creation.endswith("preplot"):
        fig, axes_idx = containing_file.plot_synthesis_status(
            synthesis_object,
            included_plots=included_plots,
            fig=fig,
            axes_idx=axes_idx,
            **plot_kwargs,
        )
    if func == "animate":
        # animate closes the matplotlib figure itself, so don't need to do it here
        path = tmp_path / "test_anim.html"
        containing_file.animate(
            synthesis_object,
            fig=fig,
            axes_idx=axes_idx,
            included_plots=included_plots,
            **plot_kwargs,
        ).save(path)
    plt.close(fig)


class TestMADDisplay:
    @pytest.fixture(scope="class", params=["rgb", "grayscale"])
    def synthesized_mad(self, request):
        # make the images really small so nothing takes as long
        if request.param == "rgb":
            img = po.load_images(IMG_DIR / "256" / "color_wheel.jpg", False).to(DEVICE)
            img = img[..., :16, :16]
        else:
            img = po.load_images(IMG_DIR / "256" / "nuts.pgm").to(DEVICE)
            img = img[..., :16, :16]

        # to serve as a metric, need to return a single value, but SSIM and MSE
        # will return a separate value for each RGB channel. Additionally, MAD
        # requires metrics are *dis*-similarity metrics, so that they return 0
        # if two images are identical (SSIM normally returns 1)
        def rgb_ssim(*args, **kwargs):
            return 1 - po.metric.ssim(*args, **kwargs).mean()

        def rgb_mse(*args, **kwargs):
            return po.metric.mse(*args, **kwargs).mean()

        mad = po.synth.MADCompetition(
            img, rgb_mse, rgb_ssim, "min", metric_tradeoff_lambda=0.1
        )
        mad.synthesize(max_iter=2, store_progress=True)
        return mad

    @pytest.fixture(
        scope="class", params=["rgb-5", "rgb-4", "grayscale-5", "grayscale-4"]
    )
    def synthesized_mad_store_progress(self, request):
        # make the images really small so nothing takes as long
        img, max_iter = request.param.split("-")
        if img == "rgb":
            img = po.load_images(IMG_DIR / "256" / "color_wheel.jpg", False).to(DEVICE)
            img = img[..., :16, :16]
        else:
            img = po.load_images(IMG_DIR / "256" / "nuts.pgm").to(DEVICE)
            img = img[..., :16, :16]
        img = torch.cat([img, img])

        # to serve as a metric, need to return a single value, but SSIM and MSE
        # will return a separate value for each RGB channel. Additionally, MAD
        # requires metrics are *dis*-similarity metrics, so that they return 0
        # if two images are identical (SSIM normally returns 1)
        def rgb_ssim(*args, **kwargs):
            return 1 - po.metric.ssim(*args, **kwargs).mean()

        def rgb_mse(*args, **kwargs):
            return po.metric.mse(*args, **kwargs).mean()

        mad = po.synth.MADCompetition(
            img, rgb_mse, rgb_ssim, "min", metric_tradeoff_lambda=0.1
        )
        mad.synthesize(max_iter=int(max_iter), store_progress=2)
        return mad

    @pytest.mark.parametrize("iteration", [None, 1, -1])
    @pytest.mark.parametrize("display_mad", [True, False])
    @pytest.mark.parametrize("loss", [True, False])
    @pytest.mark.parametrize("pixel_values", [True, False])
    @pytest.mark.parametrize(
        "fig_creation", ["auto", "auto-ratios", "pass-with", "pass-without"]
    )
    @pytest.mark.filterwarnings(
        "ignore:SSIM was designed for grayscale images:UserWarning"
    )
    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    def test_all_plot(
        self,
        synthesized_mad,
        iteration,
        display_mad,
        loss,
        pixel_values,
        fig_creation,
    ):
        # tests whether we can plot all possible combinations of plots.
        # test_custom_fig tests whether these animate correctly.
        template_test_synthesis_all_plot(
            synthesized_mad,
            iteration,
            display_mad,
            loss,
            False,
            pixel_values,
            fig_creation,
        )

    @pytest.mark.parametrize("func", ["plot", "animate"])
    @pytest.mark.parametrize(
        "fig_creation",
        [
            "custom",
            "custom-misc",
            "custom-without",
            "custom-extra",
            "custom-preplot",
        ],
    )
    @pytest.mark.filterwarnings(
        "ignore:SSIM was designed for grayscale images:UserWarning"
    )
    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    def test_custom_fig(self, synthesized_mad, func, fig_creation, tmp_path):
        # tests whether we can create our own figure and pass it to
        # MADCompetition's plotting and animating functions, specifying some or
        # all of the locations for the plots
        template_test_synthesis_custom_fig(
            synthesized_mad, func, fig_creation, tmp_path
        )

    @pytest.fixture(scope="class")
    def all_mad(self):
        # run synthesis for all 4 MAD images.
        img = po.load_images(IMG_DIR / "256" / "nuts.pgm").to(DEVICE)
        img = img[..., :16, :16]
        model1 = po.metric.mse

        # MAD requires metrics are *dis*-similarity metrics, so that they
        # return 0 if two images are identical (SSIM normally returns 1)
        def model2(*args):
            return 1 - po.metric.ssim(*args).mean()

        mad = po.synth.MADCompetition(
            img, model1, model2, "min", metric_tradeoff_lambda=1
        )
        mad.synthesize(max_iter=2)
        mad2 = po.synth.MADCompetition(
            img, model1, model2, "max", metric_tradeoff_lambda=1
        )
        mad2.synthesize(max_iter=2)
        mad3 = po.synth.MADCompetition(
            img, model2, model1, "min", metric_tradeoff_lambda=1
        )
        mad3.synthesize(max_iter=2)
        mad4 = po.synth.MADCompetition(
            img, model2, model1, "max", metric_tradeoff_lambda=1
        )
        mad4.synthesize(max_iter=2)
        return mad, mad2, mad3, mad4

    @pytest.mark.parametrize("func", ["loss", "image"])
    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    def test_helper_funcs(self, all_mad, func):
        if func == "loss":
            func = po.synth.mad_competition.plot_loss_all
        elif func == "image":
            func = po.synth.mad_competition.display_mad_image_all
        fig = func(*all_mad)
        plt.close(fig)

    @pytest.mark.parametrize("func", ["plot", "animate"])
    # plot_representation_error is an allowed value for metamer, but not MAD.
    # the second is just a typo
    @pytest.mark.parametrize("val", ["plot_representation_error", "plot_mad_image"])
    @pytest.mark.parametrize("variable", ["included_plots", "width_ratios", "axes_idx"])
    @pytest.mark.filterwarnings(
        "ignore:SSIM was designed for grayscale images:UserWarning"
    )
    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    def test_allowed_plots_exception(self, synthesized_mad, func, val, variable):
        if func == "plot":
            func = po.synth.mad_competition.plot_synthesis_status
        elif func == "animate":
            func = po.synth.mad_competition.animate
        kwargs = {}
        if variable == "included_plots":
            kwargs["included_plots"] = [val, "plot_loss"]
        elif variable == "width_ratios":
            kwargs["width_ratios"] = {val: 1, "plot_loss": 1}
        elif variable == "axes_idx":
            kwargs["axes_idx"] = {val: 0, "plot_loss": 1}
        with pytest.raises(ValueError, match=f"{variable} contained value"):
            func(synthesized_mad, **kwargs)
        plt.close()

    @pytest.mark.parametrize("iteration", [None, 0, -1, -10, 10, 2, 3, 4])
    @pytest.mark.parametrize("batch_idx", [0, 1])
    @pytest.mark.filterwarnings(
        "ignore:SSIM was designed for grayscale images:UserWarning"
    )
    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    @pytest.mark.filterwarnings("ignore:loss iteration and iteration for:UserWarning")
    def test_iteration(self, synthesized_mad_store_progress, iteration, batch_idx):
        included_plots = [
            "display_mad_image",
            "plot_loss",
            "plot_pixel_values",
        ]
        max_iter = len(synthesized_mad_store_progress.losses) - 1
        if iteration in [-10, 10]:
            expectation = pytest.raises(IndexError, match="iteration.*out of bounds")
        else:
            expectation = does_not_raise()
            if max_iter == 5:
                mad_iter = {None: 3, 0: 0, -1: 2, 2: 1, 3: 2, 4: 2}[iteration]
            elif max_iter == 4:
                mad_iter = {None: 2, 0: 0, -1: 2, 2: 1, 3: 2, 4: 2}[iteration]
        with expectation:
            fig, _ = po.synth.mad_competition.plot_synthesis_status(
                synthesized_mad_store_progress,
                iteration=iteration,
                batch_idx=batch_idx,
                included_plots=included_plots,
            )
            axes_img = fig.axes[0].images[0].get_array().data
            plt.close(fig)
            img = synthesized_mad_store_progress.saved_mad_image[mad_iter, batch_idx]
            img = einops.rearrange(po.to_numpy(img), "c h w -> h w c").squeeze()
            # rgb images are clipped while plotting
            if img.shape[-1] == 3:
                img = img.clip(0, 1)
            assert np.equal(axes_img, img).all(), "wrong saved mad_image plotted!"


class TestMetamerDisplay:
    @pytest.fixture(scope="class", params=["rgb", "grayscale"])
    def synthesized_met_nostore(self, request):
        img = request.param
        # make the images really small so nothing takes as long
        if img == "rgb":
            img = po.load_images(IMG_DIR / "256" / "color_wheel.jpg", False).to(DEVICE)
            img = img[..., :16, :16]
        else:
            img = po.load_images(IMG_DIR / "256" / "nuts.pgm").to(DEVICE)
            img = img[..., :16, :16]

        #  height=1 and order=0 to limit the time this takes, and then we
        #  only return one of the tensors so that everything is easy for
        #  plotting code to figure out (if we downsampled and were on an
        #  RGB image, we'd have a tensor of shape [1, 9, h, w], because
        #  we'd have the residuals and one filter output for each channel,
        #  and our code doesn't know how to handle that)
        class SPyr(po.simul.SteerablePyramidFreq):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def forward(self, *args, **kwargs):
                return super().forward(*args, **kwargs)[(0, 0)]

        model = SPyr(img.shape[-2:], height=1, order=1).to(DEVICE)
        met = po.synth.Metamer(img, model)
        met.synthesize(max_iter=2, store_progress=False)
        return met

    @pytest.fixture(scope="class", params=["rgb", "grayscale"])
    def synthesized_met(self, request):
        img = request.param
        # make the images really small so nothing takes as long
        if img == "rgb":
            img = po.load_images(IMG_DIR / "256" / "color_wheel.jpg", False).to(DEVICE)
            img = img[..., :16, :16]
        else:
            img = po.load_images(IMG_DIR / "256" / "nuts.pgm").to(DEVICE)
            img = img[..., :16, :16]

        #  height=1 and order=0 to limit the time this takes, and then we
        #  only return one of the tensors so that everything is easy for
        #  plotting code to figure out (if we downsampled and were on an
        #  RGB image, we'd have a tensor of shape [1, 9, h, w], because
        #  we'd have the residuals and one filter output for each channel,
        #  and our code doesn't know how to handle that)
        class SPyr(po.simul.SteerablePyramidFreq):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def forward(self, *args, **kwargs):
                return super().forward(*args, **kwargs)[(0, 0)]

        model = SPyr(img.shape[-2:], height=1, order=1).to(DEVICE)
        met = po.synth.Metamer(img, model)
        met.synthesize(max_iter=2, store_progress=True)
        return met

    @pytest.fixture(
        scope="class", params=["rgb-5", "rgb-4", "grayscale-5", "grayscale-4"]
    )
    def synthesized_met_store_progress(self, request):
        img, max_iter = request.param.split("-")
        # make the images really small so nothing takes as long
        if img == "rgb":
            img = po.load_images(IMG_DIR / "256" / "color_wheel.jpg", False).to(DEVICE)
            img = img[..., :16, :16]
        else:
            img = po.load_images(IMG_DIR / "256" / "nuts.pgm").to(DEVICE)
            img = img[..., :16, :16]
        img = torch.cat([img, img])

        #  height=1 and order=0 to limit the time this takes, and then we
        #  only return one of the tensors so that everything is easy for
        #  plotting code to figure out (if we downsampled and were on an
        #  RGB image, we'd have a tensor of shape [1, 9, h, w], because
        #  we'd have the residuals and one filter output for each channel,
        #  and our code doesn't know how to handle that)
        class SPyr(po.simul.SteerablePyramidFreq):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def forward(self, *args, **kwargs):
                return super().forward(*args, **kwargs)[(0, 0)]

        model = SPyr(img.shape[-2:], height=1, order=1).to(DEVICE)
        met = po.synth.Metamer(img, model)
        met.synthesize(max_iter=int(max_iter), store_progress=2)
        return met

    # mix together func and iteration, because iteration doesn't make sense to
    # pass to animate
    @pytest.mark.parametrize("iteration", [None, 1, -1])
    @pytest.mark.parametrize("display_metamer", [True, False])
    @pytest.mark.parametrize("loss", [True, False])
    @pytest.mark.parametrize("representation_error", [True, False])
    @pytest.mark.parametrize("pixel_values", [True, False])
    @pytest.mark.parametrize(
        "fig_creation", ["auto", "auto-ratios", "pass-with", "pass-without"]
    )
    def test_all_plot(
        self,
        synthesized_met,
        iteration,
        display_metamer,
        loss,
        representation_error,
        pixel_values,
        fig_creation,
    ):
        # tests whether we can plot all possible combinations of plots.
        # test_custom_fig tests whether these animate correctly.
        template_test_synthesis_all_plot(
            synthesized_met,
            iteration,
            display_metamer,
            loss,
            representation_error,
            pixel_values,
            fig_creation,
        )

    @pytest.mark.parametrize("func", ["plot", "animate"])
    @pytest.mark.parametrize(
        "fig_creation",
        [
            "custom",
            "custom-misc",
            "custom-without",
            "custom-extra",
            "custom-preplot",
        ],
    )
    @pytest.mark.filterwarnings(
        "ignore:Looks like representation is image-like, haven't:UserWarning"
    )
    def test_custom_fig(self, synthesized_met, func, fig_creation, tmp_path):
        # tests whether we can create our own figure and pass it to Metamer's
        # plotting and animating functions, specifying some or all of the
        # locations for the plots
        template_test_synthesis_custom_fig(
            synthesized_met, func, fig_creation, tmp_path
        )

    # mix together func and iteration, because iteration doesn't make sense to
    # pass to animate
    @pytest.mark.parametrize("display_metamer", [True, False])
    @pytest.mark.parametrize("loss", [True, False])
    @pytest.mark.parametrize("representation_error", [True, False])
    @pytest.mark.parametrize("pixel_values", [True, False])
    @pytest.mark.parametrize(
        "fig_creation", ["auto", "auto-ratios", "pass-with", "pass-without"]
    )
    def test_all_plot_nostore(
        self,
        synthesized_met_nostore,
        display_metamer,
        loss,
        representation_error,
        pixel_values,
        fig_creation,
    ):
        # tests whether we can plot all possible combinations of plots.
        # test_custom_fig tests whether these animate correctly.
        template_test_synthesis_all_plot(
            synthesized_met_nostore,
            iteration=None,
            display_synth=display_metamer,
            loss=loss,
            representation_error=representation_error,
            pixel_values=pixel_values,
            fig_creation=fig_creation,
        )

    @pytest.mark.parametrize("func", ["plot"])
    @pytest.mark.parametrize(
        "fig_creation",
        [
            "custom",
            "custom-misc",
            "custom-without",
            "custom-extra",
            "custom-preplot",
        ],
    )
    @pytest.mark.filterwarnings(
        "ignore:Looks like representation is image-like, haven't:UserWarning"
    )
    def test_custom_fig_nostore(
        self, synthesized_met_nostore, func, fig_creation, tmp_path
    ):
        # tests whether we can create our own figure and pass it to Metamer's
        # plotting and animating functions, specifying some or all of the
        # locations for the plots
        template_test_synthesis_custom_fig(
            synthesized_met_nostore, func, fig_creation, tmp_path
        )

    @pytest.mark.parametrize("func", ["plot", "animate"])
    # display_mad_image is an allowed value for MAD but not metamer.
    # the second is just a typo
    @pytest.mark.parametrize("val", ["display_mad_image", "plot_metamer"])
    @pytest.mark.parametrize("variable", ["included_plots", "width_ratios", "axes_idx"])
    def test_allowed_plots_exception(self, synthesized_met, func, val, variable):
        if func == "plot":
            func = po.synth.metamer.plot_synthesis_status
        elif func == "animate":
            func = po.synth.metamer.animate
        kwargs = {}
        if variable == "included_plots":
            kwargs["included_plots"] = [val, "plot_loss"]
        elif variable == "width_ratios":
            kwargs["width_ratios"] = {val: 1, "plot_loss": 1}
        elif variable == "axes_idx":
            kwargs["axes_idx"] = {val: 0, "plot_loss": 1}
        with pytest.raises(ValueError, match=f"{variable} contained value"):
            func(synthesized_met, **kwargs)
        plt.close()

    @pytest.mark.parametrize("iteration", [None, 0, -1, -10, 10, 2, 3, 4])
    @pytest.mark.parametrize("batch_idx", [0, 1])
    @pytest.mark.filterwarnings("ignore:loss iteration and iteration for:UserWarning")
    def test_iteration(self, synthesized_met_store_progress, iteration, batch_idx):
        included_plots = [
            "display_metamer",
            "plot_loss",
            "plot_representation_error",
            "plot_pixel_values",
        ]
        max_iter = len(synthesized_met_store_progress.losses) - 1
        if iteration in [-10, 10]:
            expectation = pytest.raises(IndexError, match="iteration.*out of bounds")
        else:
            expectation = does_not_raise()
            if max_iter == 5:
                met_iter = {None: 3, 0: 0, -1: 2, 2: 1, 3: 2, 4: 2}[iteration]
            elif max_iter == 4:
                met_iter = {None: 2, 0: 0, -1: 2, 2: 1, 3: 2, 4: 2}[iteration]
        with expectation:
            fig, _ = po.synth.metamer.plot_synthesis_status(
                synthesized_met_store_progress,
                iteration=iteration,
                batch_idx=batch_idx,
                included_plots=included_plots,
            )
            axes_img = fig.axes[0].images[0].get_array().data
            plt.close(fig)
            img = synthesized_met_store_progress.saved_metamer[met_iter, batch_idx]
            img = einops.rearrange(po.to_numpy(img), "c h w -> h w c").squeeze()
            # rgb images are clipped while plotting
            if img.shape[-1] == 3:
                img = img.clip(0, 1)
            assert np.equal(axes_img, img).all(), "wrong saved metamer plotted!"



"""Helper functions for loading tensors, converting dtypes, etc."""
# numpydoc ignore=ES01

import contextlib
import pathlib
import warnings
from collections.abc import Callable
from typing import Literal

import imageio.v3 as iio
import numpy as np
import torch
from skimage import color
from torch import Tensor

NUMPY_TO_TORCH_TYPES = {
    bool: torch.bool,  # np.bool deprecated in fav of built-in
    np.uint8: torch.uint8,
    np.int8: torch.int8,
    np.int16: torch.int16,
    np.int32: torch.int32,
    np.int64: torch.int64,
    np.float16: torch.float16,
    np.float32: torch.float32,
    np.float64: torch.float64,
    np.complex64: torch.complex64,
    np.complex128: torch.complex128,
}

TORCH_TO_NUMPY_TYPES = {value: key for (key, value) in NUMPY_TO_TORCH_TYPES.items()}


def to_numpy(x: Tensor | np.ndarray, squeeze: bool = False) -> np.ndarray:
    r"""
    Cast tensor to numpy in the most conservative way possible.

    In order this will:

    - Detach the tensor.

    - Move it to the CPU.

    - Convert it to numpy (calling ``.numpy()``).

    - Convert to the corresponding datatype.

    Parameters
    ----------
    x
        Tensor to be converted to :class:`numpy.ndarray` on CPU. If already an array, do
        nothing.
    squeeze
        Whether to remove all dummy dimensions from input.

    Returns
    -------
    array
        The former tensor, now an array.
    """
    with contextlib.suppress(AttributeError):
        # if this fails, it's already a numpy array
        x = x.detach().cpu().numpy().astype(TORCH_TO_NUMPY_TYPES[x.dtype])
    if squeeze:
        x = x.squeeze()
    return x


def load_images(
    paths: str | list[str] | pathlib.Path | list[pathlib.Path],
    as_gray: bool = True,
    sorted_key: None | Callable = None,
) -> Tensor:
    r"""
    Load in images.

    Our models and synthesis methods generally expect their inputs to
    be 4d float32 images: ``(batch, channel, height, width)``, where the batch
    dimension contains multiple images and channel contains something
    like RGB or color channel. This function helps you get your inputs
    into that format. It accepts either a single file, a list of files,
    or a single directory containing images, will load them in,
    normalize them to lie between 0 and 1, convert them to float32,
    optionally convert them to grayscale, make them tensors, and get
    them into the right shape.

    Parameters
    ----------
    paths
        A str or list of strs. If a list, must contain paths of image
        files. If a str, can either be the path of a single image file
        or of a single directory. If a directory, we try to load every
        file it contains (using imageio.imread) and skip those we
        cannot (thus, for efficiency you should not point this to a
        directory with lots of non-image files). This is NOT recursive.
    as_gray
        Whether to convert the images into grayscale or not after
        loading them. If ``False``, we do nothing. If ``True``, we call
        skimage.color.rgb2gray on them, which will result in a single
        channel.
    sorted_key
        How to sort the images. If ``None`` and ``paths`` is a directory,
        will sort the paths alphabetically. If ``paths`` is a list of files,
        must be ``None`` and is ignored. See :ref:`python:sortinghowto`
        for details on other possible values, and note that the objects to sort
        are :class:`pathlib.Path` objects.

    Returns
    -------
    images
        4d tensor containing the images.

    Raises
    ------
    FileNotFoundError
        If any of the explicit image paths do not exist.
    ValueError
        If the images we attempt to load are not all the same shape.
    ValueError
        If ``paths`` is a single file or list of files and ``sorted_key`` is not
        ``None``.

    Warns
    -----
    UserWarning
        If ``paths`` is a directory and any of the files it contains
        are non-images.

    Examples
    --------
    When ``sorted_key=None``, images from a directory are sorted alphabetically
    by filename.

    .. plot::

      >>> import plenoptic as po
      >>> from plenoptic.data.fetch import fetch_data
      >>> img_dir = fetch_data("test_images.tar.gz") / "256"
      >>> titles = ["color_wheel", "curie", "einstein", "metal", "nuts"]
      >>> imgs = po.load_images(img_dir)
      >>> po.imshow(imgs, title=titles)
      <PyrFigure size ... with 5 Axes>

    Sort the images by the second letter of their filename:

    .. plot::

      >>> import plenoptic as po
      >>> from plenoptic.data.fetch import fetch_data
      >>> img_dir = fetch_data("test_images.tar.gz") / "256"
      >>> titles = ["metal", "einstein", "color_wheel", "curie", "nuts"]
      >>> imgs = po.load_images(img_dir, sorted_key=lambda x: x.name[1])
      >>> po.imshow(imgs, title=titles)
      <PyrFigure size ... with 5 Axes>
    """
    try:
        paths = pathlib.Path(paths)
        if paths.is_dir():
            paths = sorted(paths.iterdir(), key=sorted_key)
        else:
            paths = [paths]
            if sorted_key is not None:
                raise ValueError(
                    "When paths argument is a single file, sorted_key must be None!"
                )

    except TypeError:
        # assume it is an iterable of paths
        paths = [pathlib.Path(p) for p in paths]
        if sorted_key is not None:
            raise ValueError(
                "When paths argument is a list of paths, sorted_key must be None!"
            )

    images = []
    for p in paths:
        if not p.exists():
            raise FileNotFoundError(f"File {p} not found!")
        try:
            # this makes sure we close the file properly on except
            with open(p, "rb") as f:
                im = iio.imread(f)
        except (ValueError, OSError):
            warnings.warn(
                f"Unable to load in file {p}, it's probably not an image, skipping..."
            )
            continue
        # make it a float32 array with values between 0 and 1
        im = im / np.iinfo(im.dtype).max
        if im.ndim > 2:
            if as_gray:
                # From scikit-image 0.19 on, it will treat 2d signals as 1d
                # images with 3 channels, so only call rgb2gray when it's more
                # than 2d
                try:
                    im = color.rgb2gray(im)
                except ValueError:
                    # then maybe this is an rgba image instead
                    im = color.rgb2gray(color.rgba2rgb(im))
            else:
                # RGB(A) dimension ends up on the last one, so we rearrange
                im = np.moveaxis(im, -1, 0)
        elif im.ndim == 2 and not as_gray:
            # then expand this grayscale image to the rgb representation
            im = np.expand_dims(im, 0).repeat(3, 0)
        images.append(im)
    if len(set([i.shape for i in images])) > 1:
        raise ValueError(
            "All images must be the same shape but got the following: "
            f"{[i.shape for i in images]}"
        )
    if not images:
        paths = [p.name for p in paths]
        raise ValueError(f"None of the files found at {paths} were images!")
    images = torch.as_tensor(np.array(images), dtype=torch.float32)
    if as_gray:
        if images.ndimension() != 3:
            raise ValueError(
                "For loading in images as grayscale, this should be a 3d tensor!"
            )
        images = images.unsqueeze(1)
    else:
        if images.ndimension() == 3:
            # either this was a single color image:
            # so add the batch dimension
            #  or multiple grayscale images:
            # so add channel dimension
            images = images.unsqueeze(0) if len(paths) > 1 else images.unsqueeze(1)
    if images.ndimension() != 4:
        raise ValueError(
            "Somehow ended up with other than 4 dimensions! Not sure how we got here"
        )
    return images


def convert_float_to_int(
    image: np.ndarray, dtype: Literal[np.uint8, np.uint16] = np.uint8
) -> np.ndarray:
    r"""
    Convert numpy array from float to 8 or 16 bit integer.

    We work with float images that lie between 0 and 1, but for saving them (either as
    png or in a numpy array), we typically want to convert them to 8 or 16 bit integers.
    This function does that by multiplying it by the max value for the target dtype (255
    for 8 bit 65535 for 16 bit) and then converting it to the proper type.

    Parameters
    ----------
    image
        The image to convert, with max less than or equal to 1.
    dtype
        The target data type.

    Returns
    -------
    image
        The converted image, now with specified dtype.

    Raises
    ------
    ValueError
        If ``image`` max is greater than 1.
    """
    if image.max() > 1:
        raise ValueError(
            f"all values of image must lie between 0 and 1, but max is {image.max()}"
        )
    return (image * np.iinfo(dtype).max).astype(dtype)


def polar_radius(
    size: int | tuple[int, int],
    exponent: float = 1.0,
    origin: int | tuple[int, int] | None = None,
    device: str | torch.device | None = None,
) -> Tensor:
    """
    Make distance-from-origin (r) matrix.

    Compute a matrix of given size containing samples of a radial ramp
    function, raised to given exponent, centered at given origin.

    Parameters
    ----------
    size
        If an int, we assume the image should be of dimensions ``(size,
        size)``. if a tuple, must be a 2-tuple of ints specifying the
        dimensions.
    exponent
        The exponent of the radial ramp function.
    origin
        The center of the image. if an int, we assume the origin is at
        ``(origin, origin)``. if a tuple, must be a 2-tuple of ints
        specifying the origin (where ``(0, 0)`` is the upper left).  if
        ``None``, we assume the origin lies at the center of the matrix,
        ``(size+1)/2``.
    device
        The device to create this tensor on.

    Returns
    -------
    radius
        The polar radius matrix.
    """
    if not hasattr(size, "__iter__"):
        size = (size, size)

    if origin is None:
        origin = ((size[0] + 1) / 2.0, (size[1] + 1) / 2.0)
    elif not hasattr(origin, "__iter__"):
        origin = (origin, origin)

    xramp, yramp = torch.meshgrid(
        torch.arange(1, size[1] + 1, device=device) - origin[1],
        torch.arange(1, size[0] + 1, device=device) - origin[0],
        indexing="xy",
    )

    if exponent <= 0:
        # zero to a negative exponent raises:
        # ZeroDivisionError: 0.0 cannot be raised to a negative power
        r = xramp**2 + yramp**2
        radius = np.power(r, exponent / 2.0, where=(r != 0))
    else:
        radius = (xramp**2 + yramp**2) ** (exponent / 2.0)
    return radius


def polar_angle(
    size: int | tuple[int, int],
    phase: float = 0.0,
    origin: int | tuple[float, float] | None = None,
    direction: Literal["clockwise", "counter-clockwise"] = "clockwise",
    device: torch.device | None = None,
) -> Tensor:
    """
    Make polar angle matrix (in radians).

    Compute a matrix of given size containing samples of the polar angle (in radians,
    increasing in user-defined direction from the X-axis, ranging from -pi to pi),
    relative to given phase, about the given origin pixel.

    Parameters
    ----------
    size
        If an int, we assume the image should be of dimensions ``(size, size)``. If a
        tuple, must be a 2-tuple of ints specifying the dimensions.
    phase
        The phase of the polar angle function (in radians, clockwise from the X-axis).
    origin
        The center of the image. if an int, we assume the origin is at
        ``(origin, origin)``. if a tuple, must be a 2-tuple of ints specifying the
        origin (where ``(0, 0)`` is the upper left). If ``None``, we assume the origin
        lies at the center of the matrix, ``(size+1)/2``.
    direction
        Whether the angle increases in a clockwise or counter-clockwise direction from
        the x-axis. The standard mathematical convention is to increase
        counter-clockwise, so that 90 degrees corresponds to the positive y-axis.
    device
        The device to create this tensor on.

    Returns
    -------
    res
        The polar angle matrix.

    Raises
    ------
    ValueError
        If ``direction`` takes an illegal value.
    """
    if direction not in ["clockwise", "counter-clockwise"]:
        raise ValueError(
            "direction must be one of {'clockwise', 'counter-clockwise'}, "
            f"but received {direction}!"
        )

    if not hasattr(size, "__iter__"):
        size = (size, size)

    if origin is None:
        origin = ((size[0] + 1) / 2.0, (size[1] + 1) / 2.0)
    elif not hasattr(origin, "__iter__"):
        origin = (origin, origin)

    xramp, yramp = torch.meshgrid(
        torch.arange(1, size[1] + 1, device=device) - origin[1],
        torch.arange(1, size[0] + 1, device=device) - origin[0],
        indexing="xy",
    )
    if direction == "counter-clockwise":
        yramp = torch.flip(yramp, [0])

    res = torch.atan2(yramp, xramp)

    res = ((res + (np.pi - phase)) % (2 * np.pi)) - np.pi

    return res


def _find_min_int(vals: list[int]) -> int:
    """
    Find the minimum non-negative int not in an iterable.

    Parameters
    ----------
    vals
        Iterable(s) of ints.

    Returns
    -------
    min_idx
        Minimum non-negative int.
    """  # numpydoc ignore=ES01
    flat_vals = []
    for v in vals:
        try:
            flat_vals.extend(v)
        except TypeError:
            flat_vals.append(v)
    flat_vals = set(flat_vals)
    try:
        poss_vals = set(np.arange(max(flat_vals) + 1))
    except ValueError:
        # then this is empty sequence and thus we should return 0
        return 0
    try:
        min_int = min(poss_vals - flat_vals)
    except ValueError:
        min_int = max(flat_vals) + 1
    return min_int


def _check_tensor_equality(
    x: Tensor,
    y: Tensor,
    xname: str = "x",
    yname: str = "y",
    rtol: float = 1e-5,
    atol: float = 1e-8,
    error_prepend_str: str = "Different {error_type}",
    error_append_str: str = "",
):
    """
    Check two tensors for equality: device, size, dtype, and values.

    Raises a ValueError (with informative error messages) if any of the above are not
    equal.

    Parameters
    ----------
    x, y
        The tensors to compare.
    xname, yname
        Names of the tensors, used in error messages.
    rtol, atol
        Relative and absolute tolerance for value comparison, passed to
        :func:`torch.allclose`.
    error_prepend_str
        String to start error message with, should contain the string-formatting field
        ``"{error_type}"``.
    error_append_str
        String to finish error message with.

    Raises
    ------
    ValueError
        If any of the device, dtype, size, or values differ.
    """
    error_str = (
        f"{error_prepend_str}"
        f"\n{xname}: {{xvalue}}"
        f"\n{yname}: {{yvalue}}"
        f"{{difference}}"
        f"{error_append_str}"
    )
    if x.device != y.device:
        error_str = error_str.format(
            error_type="device", xvalue=x.device, yvalue=y.device, difference=""
        )
        raise ValueError(error_str)
    # they're allowed to be different shapes if they both have 1 element (e.g., a scalar
    # and a 1-element tensor)
    if x.shape != y.shape and not (x.nelement() == y.nelement() == 1):
        error_str = error_str.format(
            error_type="shape", xvalue=x.shape, yvalue=y.shape, difference=""
        )
        raise ValueError(error_str)
    elif x.dtype != y.dtype:
        error_str = error_str.format(
            error_type="dtype", xvalue=x.dtype, yvalue=y.dtype, difference=""
        )
        raise ValueError(error_str)
    elif not torch.allclose(x, y, rtol=rtol, atol=atol):
        error_str = error_str.format(
            error_type="values", xvalue=x, yvalue=y, difference=f"\nDifference: {x - y}"
        )
        raise ValueError(error_str)
